# -*- coding: utf-8 -*-

import logging
import copy
import itertools
from artixds.impact import Impact, ImpactObjectCheck, ImpactObjectPosition, ImpactGroup, ImpactMode
from artixds.objectdiscount import CheckObject, PositionObject, Kit, LegacyObject, SimpleObject
from artixds.roster import Roster
from artixds.campaign import DiscountCampaign
from artixds.kitlogic.kitlogic import KitLogic
from artixds.kitlogic.kitobject import KitObject
from artixds.conditionfunction import ConditionFunction
from artixds.discountmarks import PositionMark

log = logging.getLogger("Discount")


class Discount(object):
    """ Discount """
    gifts = []
    coupons = []
    counters = []
    reports = []
    discountMarks = []

    def __init__(self, campaign, objects=None, conditions=None, discountRate=None, id=None, name=None,
                 cashText=None, clientText=None, clientDisplayText=None, showCashTextToConsultant=False,
                 minPriceIgnored=False, active=True, coupons=None, counters=None, gifts=None, reports=None, discountMarks=None):
        """
        campaign - parent DicountCampaign
        objects[] - list of Object
        conditions[] - list of Condition
        discountRate - DiscountRate
        """
        self.campaign = campaign
        self.objects = objects
        self.conditions = conditions or []
        self.discountRate = discountRate
        self.id = id
        self.name = name
        self.cashText = cashText
        self.clientText = clientText
        self.clientDisplayText = clientDisplayText
        self.showCashTextToConsultant = showCashTextToConsultant
        self.minPriceIgnored = minPriceIgnored
        self.active = active
        self.coupons = coupons or []
        self.counters = counters or []
        self.gifts = gifts or []
        self.reports = reports or []
        self.discountMarks = discountMarks or []

    def isActive(self):
        return not hasattr(self, 'active') or self.active

    def isValidKitObject(self, checkObj):
        if isinstance(checkObj, Roster):
            for obj in list(filter(lambda x: type(x) is Kit, self.objects)):
                checkObj.currentKit = obj
                if checkObj.findKit(obj):
                    return True
        return False

    def presenceObjects(self):
        checkObj = any(isinstance(obj, CheckObject) for obj in self.objects)
        posObj = any(isinstance(obj, PositionObject) for obj in self.objects)
        kitNewObj = any(type(obj) is KitObject for obj in self.objects)
        kitObj = any(type(obj) is Kit for obj in self.objects)
        legObj = any(isinstance(obj, LegacyObject) for obj in self.objects)
        simpleObj = any(isinstance(obj, SimpleObject) for obj in self.objects)
        return checkObj, posObj, kitObj, legObj, kitNewObj, simpleObj

    @staticmethod
    def validate(obj, check, conditionsList, discountContext):
        """ Проверка выполнения условия """
        for condit in conditionsList:
            condition = condit
            condition.checkWrapper = check
            object = obj.data if hasattr(obj, "data") else obj.currentKit
            posnum = u"Чек - "
            if "posnum" in object:
                posnum = u"Позиция " + repr(object["posnum"]) + " - "
            # если хотя бы одно условие ложно, то завершаем расчет скидок для этой акции
            if not condition.validate(check.checkDict, object, discountContext):
                log.debug(u"        %s%s - false", posnum, condition.condition)
                return False
            log.debug(u"        %s%s - true", posnum, condition.condition)
        return True

    @staticmethod
    def isConditionsTrue(obj, check, discountContext):
        """
        Check all conditions.

        obj - current Object
        check - Check

        Возвращает 2 значения:
        1. True, если все условия выполняются, иначе False
        2. True, если какое то условие не зависящее от позиции не выполнено и дальнейший расчет бесполезен, иначе False
        """
        # расчитываем условия не зависящие от позиций
        if not discountContext.posIndependentConditionsSuccess:
            # если хотя бы одно условие ложно, то завершаем расчет скидок для этой акции
            if not Discount.validate(obj, check, discountContext.posIndependent, discountContext):
                return False, True
            discountContext.posIndependentConditionsSuccess = True

        # расчитываем условия зависящие от позиции
        return Discount.validate(obj, check, discountContext.posDependent, discountContext), False

    def injectPosition(self, check, qty, currentNumber, data, partOfDisKit):
        """
        Create new position and inject it to check
        """
        log.debug("inject position number %s", currentNumber)
        position = data
        position["posnum"] = currentNumber
        position["sumn"] = position["price"] * qty
        position["sumb"] = position["sumb"] / position["cquant"] * qty
        position["cquant"] = qty
        # TODO: add recalculation "sum"
        position["partOfDisKit"] = partOfDisKit
        check.addPosition(position)

    def getGoods(self, obj, check, code):
        res = list(filter(lambda obj: isinstance(obj, PositionObject) and obj.data["code"] == code, check.getObjects()))
        if len(res) < 0:
            raise Exception("Couldn't find goods with code %s in check" % code)
        goods = res[0]
        return goods

    def getKitImpact(self, check, obj, discountContext):
        """
        kit - Kit
        check - checkWrapper
        """
        check.isModified = True
        self.discountRate.checkWrapper = check
        impactGroup = ImpactGroup()
        kit = obj.currentKit
        entry = obj.findKit(kit)
        applyTimes = 0
        while entry and applyTimes < kit.maxOccurances(check):
            applyTimes += 1
            obj.subtractKitEntry(entry)
            entry = obj.findKit(kit)
        obj.applyRate(self.discountRate, self.campaign, self, discountContext)
        pureImpacts = obj.pureImpacts()
        structureImpacts = obj.structureImpacts()
        impactGroup.addImpacts(structureImpacts)
        impactGroup.addImpacts(pureImpacts)
        for impact in structureImpacts:
            check.branchPosition(impact.impactObject.positionNumber, impact.branchDetails)
        for posnum in obj.partsOfKit():
            check.markAsPartOfKit(posnum)
        for impact in pureImpacts:
            check.positionsImpact.add(impact.impactObject.positionNumber)

        return impactGroup, applyTimes

    def makeKit(self, originalKit, items):
        """
        Создание набора на основании @originalKit и добаление в его KitObjectItem'ы свойства 'codes'
        В первую часть набора попадёт items[0], во вторую часть набора попадёт items[1] и т.д.
        """
        kit = copy.deepcopy(originalKit)
        itemIter = iter(items)
        for kitObjectItem in kit:
            setattr(kitObjectItem, 'codes', set([next(itemIter)]))
        return kit

    def applyKitImpacts(self, kitLogic, check, impactGroup):
        """
        Применение дисконтных воздействий от логики наборов @arg kitLogic к чеку @arg check
        и добавлением дисконтных воздействий в @arg impactGroup
        """
        impactGroup.addImpacts(kitLogic.uniteImpacts)
        impactGroup.addImpacts(kitLogic.structureImpacts)
        impactGroup.addImpacts(kitLogic.pureImpacts)
        for impact in kitLogic.uniteImpacts:
            check.unitePosition(impact.impactObject.positionNumber, impact.qty, impact.uniteDetails)
        for impact in kitLogic.structureImpacts:
            check.branchPosition(impact.impactObject.positionNumber, impact.branchDetails)
        for posnum in kitLogic.partsOfKit:
            check.markAsPartOfKit(posnum)
        for impact in kitLogic.pureImpacts:
            check.positionsImpact.add(impact.impactObject.positionNumber)

    @staticmethod
    def fillCodes(kit, check):
        """
        Для каждого элемента набора(KitObjectItem) заполняет набор кодов, которые подходят по условию в эту часть набора
        """
        for kitObjectItem in kit:
            kitObjectItem.codes = set()
            for pos in check.getGoodsList():
                if hasattr(kitObjectItem, 'conditions') and kitObjectItem.conditions:
                    posDependent = []
                    for conditItem in kitObjectItem.conditions:
                        posDependentCondition = list(
                            filter(lambda x: x in conditItem.condition, ConditionFunction.posConditionList))
                        if posDependentCondition:
                            posDependent.append(conditItem)
                    if Discount.validate(pos, check, posDependent, None):
                        kitObjectItem.codes.add(pos.data['code'])
                else:
                    posCatalogs = set(catalog["catalogid"] for catalog in pos.data['tmc']['catalogs'])
                    if set(kitObjectItem.catalogCodes) & posCatalogs:
                        kitObjectItem.codes.add(pos.data['code'])

    def getKitNewImpact(self, check, obj, discountContext):
        """
        Расчет скидок на новый набор
        """
        check.isModified = True
        impactGroup = ImpactGroup()
        self.discountRate.checkWrapper = check
        kit = None

        for obj in list(filter(lambda x: type(x) is KitObject, self.objects)):
            kit = obj
            break

        applyTimes = 0
        Discount.fillCodes(kit, check)
        if hasattr(kit, 'cartesian') and kit.cartesian and (not hasattr(kit, 'byTmc') or (hasattr(kit, 'byTmc') and not kit.byTmc)):
            # пока в чеке есть наборы, то будем их выделять
            while True:
                maxOccurance = 0
                maxKits = []
                # создаём декартово произведение кодов товаров из разных частей набора
                # для каждого элемента декартова произведения имеем (code1, code2, ..., codeN),
                # где code1 - код товара, который должен быть в первой части набора,
                # code2 - код товара, который должен быть во второй части набора
                # и codeN - код товара, который должен быть в N-ой части набора
                # таким образом в каждой части набора будет только 1 товар
                for item in itertools.product(*[getattr(kitObjectItem, 'codes') for kitObjectItem in kit]):
                    # создаём набор
                    newKit = self.makeKit(kit, item)
                    kitLogic = KitLogic(check, newKit, self.discountRate, self.campaign, self, discountContext)
                    # ищем количество вхождений набора
                    occ = kitLogic.getKitOccurrences()
                    # обновляем информацию по максимальному вхождению набора и списку таких наборов
                    if occ > maxOccurance:
                        maxKits = [newKit]
                        maxOccurance = occ
                    elif occ == maxOccurance and occ > 0:
                        maxKits.append(newKit)
                # если нет вхождений набора, то выходим
                if maxOccurance == 0:
                    break

                # мап - <сумма скидок в PureImpacts, логика_набора>
                impacts = {}
                # для каждого из наборов с одинаковым максимальным вхождением рассчитываем скидку
                for kt in maxKits:
                    kLogic = KitLogic(check, kt, self.discountRate, self.campaign, self, discountContext)
                    kLogic.findKit()
                    discSum = sum(impact.absoluteValue for impact in kLogic.pureImpacts)
                    impacts[discSum] = kLogic
                # выбираем минимальный набор с минимальной скидкой
                minKit = impacts[min(impacts.keys())]
                # применяем набор и продолжаем поиск других наборов
                self.applyKitImpacts(minKit, check, impactGroup)
                applyTimes += maxOccurance
        else:
            kitLogic = KitLogic(check, kit, self.discountRate, self.campaign, self, discountContext)
            if not kitLogic.findKit():
                return impactGroup, applyTimes
            applyTimes = kitLogic.getKitOccurrences()
            self.applyKitImpacts(kitLogic, check, impactGroup)
        return impactGroup, applyTimes

    def getCheckImpact(self, check, obj, discountContext):
        checkDict = check.checkDict
        self.discountRate.checkWrapper = check

        check.isModified = True
        impactGroup = ImpactGroup()
        impact = Impact()
        impact.campaign = self.campaign
        impact.discount = self
        impact.innerObject = obj
        impact.impactObject = ImpactObjectCheck()
        card = discountContext.cards[0] if discountContext.cards else None
        impact.absoluteValue = self.discountRate.getAbsoluteValue(obj, checkDict, card)
        impact.cards = discountContext.cards
        impact.coupons = discountContext.coupons
        impact.counters = discountContext.counters

        if impact.discountMode == ImpactMode.AUTOMATIC.value:
            if discountContext.coupons:
                impact.discountMode = ImpactMode.BYCOUPON.value
            elif discountContext.cards:
                impact.discountMode = ImpactMode.BYCARD.value

        impactGroup.addImpact(impact)
        return impactGroup, 1

    def getPosImpact(self, check, obj, discountContext):
        checkDict = check.checkDict
        self.discountRate.checkWrapper = check

        check.isModified = True
        impactGroup = ImpactGroup()
        impact = Impact()
        impact.campaign = self.campaign
        impact.discount = self
        impact.innerObject = obj
        impact.impactObject = ImpactObjectPosition(positionNumber=obj.data["posnum"], code=obj.data["code"])
        impact.qty = obj.data["cquant"]
        card = discountContext.cards[0] if discountContext.cards else None
        impact.absoluteValue = self.discountRate.getAbsoluteValue(obj, checkDict, card)
        impact.cards = discountContext.cards
        impact.coupons = discountContext.coupons

        if impact.discountMode == ImpactMode.AUTOMATIC.value:
            if discountContext.coupons:
                impact.discountMode = ImpactMode.BYCOUPON.value
            elif discountContext.cards:
                impact.discountMode = ImpactMode.BYCARD.value

        impactGroup.addImpact(impact)
        check.positionsImpact.add(obj.data["posnum"])
        return impactGroup, 1

    def getLegacyImpact(self, check, obj, discountContext):
        from legacydiscount import main as legacySystem
        from artixds.branchdetail import BranchDetail

        self.discountRate.checkWrapper = check
        check.isModified = True
        impactGroup = ImpactGroup()

        b, d = legacySystem.calculateImpact(check.check)
        structureImpacts = []
        pureImpacts = []

        for i in b:
            branchDetails = []
            for sImp in b[i]:
                detail = BranchDetail(sImp['posnum'], sImp['qty'], sImp['sum'], sImp['sum'], False)
                branchDetails.append(detail)
            structureImpacts.append(Impact.branchImpact(ImpactObjectPosition(int(i), None), branchDetails))

        for impact in structureImpacts:
            check.branchPosition(impact.impactObject.positionNumber, impact.branchDetails)

        for dImp in d:
            impact = Impact()
            impact.cards = []
            impact.campaign = DiscountCampaign(id = dImp.rate.scheme.code, name = dImp.rate.scheme.name)

            code = 0
            if hasattr(dImp.innerObject, 'code'):
                code = dImp.innerObject.code
            impact.impactObject = ImpactObjectPosition(positionNumber=dImp.innerObject.posnum, code=code)
            impact.qty = dImp.innerObject.cquant
            impact.absoluteValue = float(dImp.absoluteValue)

            if dImp.card:
                if isinstance(dImp.card.dump(), dict):
                    impact.cards.append(dImp.card.dump())
                else:
                    impact.cards.append({'codeinternal' : 1})

            if dImp.discountMode == "bydisckit":
                impact.discountMode = ImpactMode.BYDISCKIT.value
                check.markAsPartOfKit(dImp.innerObject.posnum)
            elif dImp.discountMode == "bycard":
                impact.discountMode = ImpactMode.BYCARD.value
            elif dImp.discountMode == "automatic":
                impact.discountMode = ImpactMode.AUTOMATIC.value

            if impact.discountMode == ImpactMode.AUTOMATIC.value and discountContext.cards:
                impact.discountMode = ImpactMode.BYCARD.value
            pureImpacts.append(impact)

            self.discountRate.type = dImp.rate.type
            self.discountRate.value = float(dImp.rate.value)
            self.discountRate.items = []
            self.discountRate.checkWrapper = None

            impact.discount = Discount(self)
            impact.discount.id = dImp.rate.disccode
            impact.discount.name = dImp.rate.name
            impact.discount.discountRate = copy.deepcopy(self.discountRate)
            if check.checkDict.get("cashtype") != "SSCO" or self.showCashTextToConsultant:
                impact.discount.cashText = copy.deepcopy(self.cashText)
            impact.discount.clientText = copy.deepcopy(self.clientText)
            if check.checkDict.get("cashtype") == "SSCO":
                impact.discount.clientDisplayText = copy.deepcopy(self.clientDisplayText)
            impact.discount.showCashTextToConsultant = copy.deepcopy(self.showCashTextToConsultant)
            impact.discount.minPriceIgnored = copy.deepcopy(self.minPriceIgnored)
            impact.discount.active = copy.deepcopy(self.active)
            check.markPositionByDiscount(dImp.innerObject.posnum)

        impactGroup.addImpacts(structureImpacts)
        impactGroup.addImpacts(pureImpacts)
        return impactGroup, len(impactGroup.getImpacts())

    def getSimpleImpact(self, check, obj, discountContext):
        from simplediscount import main as simpleSystem
        from artixds.branchdetail import BranchDetail

        self.discountRate.checkWrapper = check
        check.isModified = True
        impactGroup = ImpactGroup()

        b, d = simpleSystem.calculateImpact(check.check)
        structureImpacts = []
        pureImpacts = []

        for i in b:
            branchDetails = []
            for sImp in b[i]:
                detail = BranchDetail(sImp['posnum'], sImp['qty'], sImp['sum'], sImp['sum'], False)
                branchDetails.append(detail)
            structureImpacts.append(Impact.branchImpact(ImpactObjectPosition(int(i), None), branchDetails))

        for impact in structureImpacts:
            check.branchPosition(impact.impactObject.positionNumber, impact.branchDetails)

        for dImp in d:
            impact = Impact()
            impact.cards = []
            impact.campaign = DiscountCampaign(id = dImp.rate.scheme.code, name = dImp.rate.scheme.name)

            code = 0
            if hasattr(dImp.innerObject, 'code'):
                code = dImp.innerObject.code
            impact.impactObject = ImpactObjectPosition(positionNumber=dImp.innerObject.posnum, code=code)
            impact.qty = dImp.innerObject.cquant
            impact.absoluteValue = float(dImp.absoluteValue)

            if dImp.card:
                if isinstance(dImp.card.dump(), dict):
                    impact.cards.append(dImp.card.dump())
                else:
                    impact.cards.append({'codeinternal': 1})

            if dImp.discountMode == "bydisckit":
                impact.discountMode = ImpactMode.BYDISCKIT.value
                check.markAsPartOfKit(dImp.innerObject.posnum)
            elif dImp.discountMode == "bycard":
                impact.discountMode = ImpactMode.BYCARD.value
            elif dImp.discountMode == "automatic":
                impact.discountMode = ImpactMode.AUTOMATIC.value

            if impact.discountMode == ImpactMode.AUTOMATIC.value and discountContext.cards:
                impact.discountMode = ImpactMode.BYCARD.value
            pureImpacts.append(impact)

            self.discountRate.type = dImp.rate.type
            self.discountRate.value = float(dImp.rate.value)
            self.discountRate.items = []
            self.discountRate.checkWrapper = None

            impact.discount = Discount(self)
            impact.discount.id = dImp.rate.disccode
            impact.discount.name = dImp.rate.name
            impact.discount.discountRate = copy.deepcopy(self.discountRate)
            if check.checkDict.get("cashtype") != "SSCO" or self.showCashTextToConsultant:
                impact.discount.cashText = copy.deepcopy(self.cashText)
            impact.discount.clientText = copy.deepcopy(self.clientText)
            if check.checkDict.get("cashtype") == "SSCO":
                impact.discount.clientDisplayText = copy.deepcopy(self.clientDisplayText)
            impact.discount.showCashTextToConsultant = copy.deepcopy(self.showCashTextToConsultant)
            impact.discount.minPriceIgnored = copy.deepcopy(self.minPriceIgnored)
            impact.discount.active = copy.deepcopy(self.active)
            check.markPositionByDiscount(dImp.innerObject.posnum)

        impactGroup.addImpacts(structureImpacts)
        impactGroup.addImpacts(pureImpacts)
        return impactGroup, len(impactGroup.getImpacts())


    def getPositionMarks(self):
        return list(filter(lambda mark: isinstance(mark, PositionMark), self.discountMarks))
