# -*- coding: utf-8 -*-

import logging
from datetime import date, datetime, time
from artixds.conditionfunction import ConditionFunction
import re

log = logging.getLogger("DiscountCondition")


class DiscountCondition(object):
    """
    Condition
    """
    def __init__(self, condition=None, description=None):
        """
        condition - python code
        description - condition explanation
        """
        self.description = description
        self.condition = condition
        self.checkWrapper = None

    def eval(self, cf, check, object):
        scope = dict(globals(), **locals())
        return eval(self.condition, scope)
        
    def validate(self, check=None, object=None, discountContext=None):
        """
        Eval condition
        obj - current Object
        check - Check
        Return result of eval condition (true or false), if error in condition- return false
        """
        try:
            cf = ConditionFunction(check, object, checkWrapper=self.checkWrapper)
            res = self.eval(cf, check, object)
        except Exception as e:
            log.error("error in eval call %s" % self.condition)
            log.exception(e)
            res = False
        log.debug("validate condition: <%s>, result=%s ", self.condition, res)
        return res

    def __repr__(self):
        return "<DiscountCondition(%s)>" % (repr(self.condition))


class DiscountCardCondition(object):
    """
    Условие скидки по дисконтной карте
    """
    def __init__(self, condition=None, description=None):
        """
        condition - строка. Выражение на языке Python, возвращающее True или False.
        В выражении доступна текущая карта под именем card.
        Пример: 'card["cardgroup"]["code"]==2'
        """
        self.description = description
        self.condition = condition
        self.checkWrapper = None
                
    def predicate(self, check, object, card):
        try:
            log.debug("testing card %s" % card)
            cf = ConditionFunction(check, object, card, checkWrapper=self.checkWrapper)
            return self.eval(cf, check, object, card)
        except Exception as e:
            log.error('error in eval call %s' % self.condition)
            log.exception(e)
        return False

    def eval(self, cf, check, object, card):
        scope = dict(globals(), **locals())
        return eval(self.condition, scope)

    def validate(self, check=None, object=None, discountContext=None):
        result = False
        try:
            cards = list(filter(lambda card: self.predicate(check, object, card), check["cardrecords"]))
            if cards:
                # добавление только тех карт, которых нет в контексте
                for card in cards:
                    if not any(c for c in discountContext.cards if c.get('codeinternal', 0) == card.get('codeinternal', 0)):
                        discountContext.cards.append(card)
                result = True
        except Exception as e:
            log.error("error in eval call %s" % self.condition)
            log.exception(e)
        log.debug("validate condition: <%s>, result=%s ", self.condition, result)
        return result

    def __repr__(self):
        return "<DiscountCondition(%s)>" % (repr(self.condition))


class DiscountCouponCondition(object):
    """
    Условие скидки по купону
    """
    def __init__(self, condition=None, description=None):
        """
        condition - строка. Выражение на языке Python, возвращающее True или False.
        В выражении доступен текущий купон под именем coupon.
        Пример: 'coupon["number"]=="2"'
        """
        self.description = description
        self.condition = condition
        self.checkWrapper = None
                
    def predicate(self, check, object, coupon):
        try:
            log.debug("testing coupon %s" % coupon)
            cf = ConditionFunction(check, object, coupon=coupon, checkWrapper=self.checkWrapper)
            return self.eval(cf, check, object, coupon)
        except Exception as e:
            log.error('error in eval call %s' % self.condition)
            log.exception(e)
        return False

    def eval(self, cf, check, object, coupon):
        scope = dict(globals(), **locals())
        return eval(self.condition, scope)

    def validate(self, check=None, object=None, discountContext=None):
        result = False
        try:
            coupons = list(filter(lambda coupon: self.predicate(check, object, coupon), check["coupons"]))
            if coupons:
                discountContext.coupons.extend(coupons)
                result = True
        except Exception as e:
            log.error("error in eval call %s" % self.condition)
            log.exception(e)
        log.debug("validate condition: <%s>, result=%s ", self.condition, result)
        return result

    def __repr__(self):
        return "<DiscountCondition(%s)>" % (repr(self.condition))
