# -*- coding: utf-8 -*-

import logging
from enum import Enum
from artixds.ratefunction import RateFunction

log = logging.getLogger("DiscountRate")


class DiscountRateType(Enum):
    PERCENT = 'percent'
    SUM = 'sum'
    INDEX = 'index'
    PRICE = 'price'


def createDiscountRate(type, value):
    return {DiscountRateType.PERCENT.value: PercentDiscountRate,
        DiscountRateType.SUM.value: SumDiscountRate,
        DiscountRateType.INDEX.value: IndexDiscountRate,
        DiscountRateType.PRICE.value: PriceDiscountRate
    }.get(type, None)(value)


class DiscountRate(object):
    def __init__(self, type=None):
        self.type = type
        self.checkWrapper = None

    def getAbsoluteValue(self, obj, check, card=None, **additionalObjects):
        """
        check - current checkWrapper
        # TODO: what's kind of obj? Different  for different DiscountRate?
        obj - ????
        """
        return 0

    def getRateType(self):
        return self.type


class PercentDiscountRate(DiscountRate):
    def __init__(self, value=None):
        super(PercentDiscountRate, self).__init__(DiscountRateType.PERCENT.value)
        self.value = value

    def __repr__(self):
        return "%s%%" % self.value

    def getAbsoluteValue(self, obj, check, card=None, **additionalObjects):
        return obj.getPercentOfFinalSum(self.value)


class SumDiscountRate(DiscountRate):
    def __init__(self, value=None):
        super(SumDiscountRate, self).__init__(DiscountRateType.SUM.value)
        self.value = value

    def getAbsoluteValue(self, obj, check, card=None, **additionalObjects):
        return obj.getAbsoluteDiscountSum(self.value)


class IndexDiscountRate(DiscountRate):
    def __init__(self, value=None):
        super(IndexDiscountRate, self).__init__(DiscountRateType.INDEX.value)
        self.value = value

    def getAbsoluteValue(self, obj, check, card=None, **additionalObjects):
        return obj.getIndexDiscountSum(int(self.value))


class PriceDiscountRate(DiscountRate):
    def __init__(self, value=None):
        super(PriceDiscountRate, self).__init__(DiscountRateType.PRICE.value)
        self.value = value

    def getAbsoluteValue(self, obj, check, card=None, **additionalObjects):
        return obj.getDiscountSumForPrice(self.value)

    def getValue(self):
        return self.value


class CalculateDiscountRate(DiscountRate):
    def __init__(self, value=None, type=DiscountRateType.PERCENT.value):
        super(CalculateDiscountRate, self).__init__(type)
        self.value = value

    def eval(self, rf, object, check, card=None, **additionalObjects):
        scope = dict(globals(), **locals())
        return eval(self.value, scope)

    def getAbsoluteValue(self, obj, check, card=None, **additionalObjects):
        try:
            log.info('calculating rate using expression: %s' % self.value)
            log.info('Current card info: %s' % card)
            rf = RateFunction(check, obj, card, self.checkWrapper)
            if hasattr(obj, "data"):
                data = obj.data
            else:
                data = obj.__dict__
            val = self.eval(rf, data, check, card, **additionalObjects)
        except Exception as e:
            log.error("error in eval call: %s" % e)
            log.exception(e)
            val = 0
        log.info('calculated value is : %s' % val)
        log.info('type is: %s' % self.type)
        return createDiscountRate(self.type, val).getAbsoluteValue(obj, check)


class KitDiscountRateItem(object):
    def __init__(self, kitItem=None, discountRate=None):
        self.kitItem = kitItem
        self.discountRate = discountRate

    def __repr__(self):
        return "KitDiscountRateItem(kitItem=%s,discountRate=%s)" % (self.kitItem, self.discountRate)


class KitDiscountRate(DiscountRate):
    def __init__(self, items=None):
        super(KitDiscountRate, self).__init__()
        self.items = items if items else []

    def __repr__(self):
        return "KitDiscountRate(items=%s)" % self.items

    def addItem(self, kitItem, discountRate):
        self.items.append(KitDiscountRateItem(kitItem, discountRate))

    def discountRateForItem(self, kitItem):
        for item in self.items:
            if item.kitItem == kitItem:
                return item.discountRate
        return None

    def getAbsoluteValues(self, kitEntryItem, card=None):
        return []
