# -*- coding: utf-8 -*-

import logging
import copy

from artixds.objectdiscount import PositionObject
from artixds.impact import Impact, ImpactType, ImpactObjectCheck, ImpactObjectPosition, ImpactMode
from artixds.branchdetail import BranchDetail
from artixds.discountrate import DiscountRateType
from config import Config

from artixds.kitlogic.klkititem import KlKitItem
from artixds.kitlogic.klposition import KlPosition
from artixds.kitlogic.klpositionpart import KlPositionPart

log = logging.getLogger("KitLogic")


class KitLogic(object):
    """
    Новая логика для работы с наборами

    При помощи метода findKit ищется набор в чеке и если набор найден возвращается true.
    После чего из полей structureImpacts, pureImpacts, uniteImpacts можно получить воздействия,
    а в поле partsOfKit список номеров позиций которые входят в набор.

    Если прям совсем непонятно как работает, вот тут немного понятнее описано
    http://docs.artix.su/x/EwbY
    """
    def __init__(self, check, kit, discRate, campaign, discount, discountContext):
        self.check = check
        self.kit = kit
        self.discRate = discRate
        self.nextPosNum = check.getNextPositionNumber()
        self.campaign = campaign
        self.discount = discount
        self.discountContext = discountContext

        self.pureImpacts = []   # Скидочные воздействия в которых нет номера позиции
        self.pureImpactsWithoutNum = []
        self.uniteImpacts = []
        self.structureImpacts = []

    def getKitOccurrences(self):
        positions = [KlPosition(pos) for pos in self.check.getGoodsList() if not pos.data["partOfKit"] and not pos.data["packingMode"]]
        if not positions:
            return 0
        return self.occ(self.createKitItems(positions))

    def createKitItems(self, positions):
        # Элементы набора
        posPerKitItem = [KlKitItem(ki, self.discRate.discountRateForItem(ki), self.check) for ki in self.kit]

        # Пересортируем элементы набора так, чтобы первыми были части со скидками
        posPerKitItem = sorted(posPerKitItem, key=lambda x: x.discRate is None)

        # Раскидываем позиции по частям набора
        # Изменения происходят внутри posPerKitItem
        self.distributeForKitItems(positions, posPerKitItem)
        return posPerKitItem

    def findKit(self):
        log.debug(u"Поиск вхождений набора '%s' в чек" % self.kit.name)

        # Обворачиваем позиции
        positions = [KlPosition(pos) for pos in self.check.getGoodsList() if not pos.data["partOfKit"] and not pos.data["packingMode"]]
        if not positions:
            return False

        if hasattr(self.kit, 'byTmc') and self.kit.byTmc:
            # получаем множество кодов товаров
            codeSet = set([p.getCode() for p in positions])

            # для каждого из кодов товара ищем наборы
            result = False
            for code in codeSet:
                result |= self.makeKit(filter(lambda p: p.getCode() == code, positions))
            return result

        return self.makeKit(positions)

    def makeKit(self, positions):
        posPerKitItem = self.createKitItems(positions)
        occurrenceCount = self.occ(posPerKitItem)
        log.debug(u"Количество вхождений набора - %s" % occurrenceCount)
        if occurrenceCount < 1:
            return False

        self.united = set()     # объединенные позиции
        self.posWithBranches = set()
        self.partsOfKit = set()

        # Выделим позиции, которые входят в набор
        self.selectPositions(occurrenceCount, posPerKitItem)

        # Создаем импакты объединения
        for pos in self.united:
            self.uniteImpacts.append(Impact.uniteImpact(
                ImpactObjectPosition(pos.posnum, pos.getCode()), pos.bquant, pos.unitedPositions))

        # Создаем структурные импакты
        # Сортируем, чтобы новые позиции были с номерами по порядку
        for pos in sorted(self.posWithBranches, key=lambda x: x.posnum):
            pos.complete(self.getNextPosNum)
            if not pos.parts:
                # Если нет частей, значит объединили
                self.partsOfKit.add(pos.posnum)
                continue
            self.structureImpacts.append(Impact.branchImpact(
                    ImpactObjectPosition(pos.posnum, pos.getCode()),
                    [BranchDetail(i.posnum, i.bquant, i.sumn, i.sumb, i.partOfKit) for i in pos.parts]))
            self.partsOfKit.update([i.posnum for i in pos.parts if i.partOfKit])

        # Не забыли про импакты с частями позиций, у которых не было номеров
        for impact, pos in self.pureImpactsWithoutNum:
            if pos.posnum is None:
                # Если нет номера, значит объединили
                impact.impactObject.positionNumber = pos.source().posnum
            else:
                impact.impactObject.positionNumber = pos.posnum

        log.debug(u"Воздействия объединения - %s" % self.uniteImpacts)
        log.debug(u"Cтруктурные воздействия - %s" % self.structureImpacts)
        log.debug(u"Скидочные воздействия - %s" % self.pureImpacts)
        log.debug(u"Номера позиций, которые попали в набор - %s" % self.partsOfKit)
        self.pureImpacts.sort(key=lambda x: x.impactObject.positionNumber)
        return True

    def getNextPosNum(self):
        """ Возвращает номер позиции для бранча """
        result = self.nextPosNum
        self.nextPosNum += 1
        return result

    def distributeForKitItems(self, positions, posPerKitItem):
        """
        Раскидать позиции по частям набора

        Если позиция подходит в оба элемента набора, позиция добавляется в каждый
        в виде KlPositionPart

        Позиции которые используются в других наборах и расфасованные игнорируются
        """
        for pos in positions:
            for i in range(len(posPerKitItem)):
                kitItem = posPerKitItem[i].kitItem
                if posPerKitItem[i].fract and not pos.fract:
                    log.warn(u"Не весовая позиция %s не может участвовать в части набора с дробным кол-вом %s" % (pos, kitItem.name))
                    continue

                if pos.pos.data['code'] in getattr(kitItem, 'codes'):
                    if not pos.usedTimes():
                        pos.kitItemNum = i
                        posPerKitItem[i].append(pos)
                    elif pos.usedTimes() > 1:
                        posPerKitItem[i].append(pos.createPart())
                    else:
                        posPerKitItem[pos.kitItemNum].remove(pos)
                        posPerKitItem[pos.kitItemNum].append(pos.createPart())
                        posPerKitItem[i].append(pos.createPart())

    def applyDisc(self, pos, kitItem, occurrenceCount, kitPrice=0.0, applyPriceDiscount=False):
        """ Применение скидочного воздействия на позицию pos """
        if kitItem.discRate:
            if ((kitItem.discRate.getRateType() != DiscountRateType.PRICE.value and applyPriceDiscount) or
                (kitItem.discRate.getRateType() == DiscountRateType.PRICE.value and not applyPriceDiscount)):
                return

        isPart = isinstance(pos, KlPositionPart)
        klPos = pos.source()

        if isPart:
            # сохраняем разбитые позиции, чтобы потом сделать структурные импакты
            self.posWithBranches.add(klPos)
        else:
            # сохраняем номера позиций, которые стали частью набора
            # у части позиции еще нет номера, потом сохраним
            self.partsOfKit.add(pos.posnum)
        pos.partOfKit = True

        if kitItem.discRate:
            rateValue = kitItem.discRate.getAbsoluteValue(PositionObject(klPos.pos.data, pos), self.check.checkDict)
            # Распределять скидку на набор по всему чеку
            distributeByCheck = hasattr(self.kit, 'distributeByCheck') and self.kit.distributeByCheck and (not hasattr(self.kit, 'byTmc') or (hasattr(self.kit, 'byTmc') and not self.kit.byTmc))

            # Если суммовая скидка
            if kitItem.discRate.getRateType() == DiscountRateType.SUM.value:
                if distributeByCheck:
                    # Если распределяем скидку по всему чеку, то выбираем минимум между суммовой скидкой и ценой набора
                    kitPrice = sum(item.sumn for item in kitItem.positions)
                    rateValue = min(rateValue, kitPrice)
                rateValue = rateValue * pos.bquant / kitItem.kitItem.quantity
            # Если воздействие Price
            if kitItem.discRate.getRateType() == DiscountRateType.PRICE.value:
                # Цена набора из воздействия Price  * количество наборов
                priceValue = kitItem.discRate.getValue() * occurrenceCount
                rateValue = round(pos.sumn - (pos.sumn * priceValue / kitPrice), 2)

            impact = Impact(type=ImpactType.PURE.value,
                campaign=self.campaign,
                discount=self.discount,
                absoluteValue=rateValue,
                impactObject=ImpactObjectCheck() if distributeByCheck else ImpactObjectPosition(pos.posnum, code=klPos.getCode()),
                qty=pos.bquant,
                mode=ImpactMode.BYDISCKIT.value,
                discountRate=kitItem.discRate)
            impact.cards = self.discountContext.cards
            impact.counters = self.discountContext.counters
            impact.coupons = self.discountContext.coupons

            self.pureImpacts.append(impact)
            if isPart:
                # у части позиции еще нет номера, потом установим
                self.pureImpactsWithoutNum.append((impact, pos))

    def distributeCommon(self, posPerKitItem):
        """ Раскидывание общих позиций между частями набора """
        if any([ki.hasCommon for ki in posPerKitItem]):
            # Считаем общее количество в позициях, которые входят в пересекающиеся части набора
            commonPosSet = set()
            commonPosCount = sum([ki.uniqueCount(commonPosSet) for ki in posPerKitItem if ki.hasCommon])
            # Считаем общее количество весов (вес это количество элементов в части набора)
            weighSum = sum([ki.kitItem.quantity for ki in posPerKitItem if ki.hasCommon])
            # коэффициент распределения общих позиций (его умножаем на кол-во в элементе набора
            # и получаем кол-во позиций для оптимального кол-ва вхождения набора)
            factor = commonPosCount / weighSum

            # Распределение общих позиций
            for kitItem in posPerKitItem:
                kitItem.distributeCommon(factor)

    def selectPositions(self, occurrenceCount, posPerKitItem):
        """
        Выделяет в каждой части набора позиции, которые попали в набор для
        заданного количества вхождений

        Может объединить позиции. Если часть позиции занимает все количество исходной,
        из нее сделает полную. От полной может отрезать часть.
        """
        unitePositionsInKit = Config().getBool('Discounts', 'unitePositionsInKit', True)
        log.debug(u"Объединение позиций " + u"включено" if unitePositionsInKit else u"выключено")

        for kitItem in posPerKitItem:
            neededCount = round(occurrenceCount * kitItem.kitItem.quantity, 3)
            # позиция, в которую объединяются другие
            collector = None
            # список позиций для применения воздействия Price для набора
            positionsForPriceDiscount = []

            for i in range(len(kitItem.positions)):
                pos = kitItem.positions[i]
                isPart = isinstance(pos, KlPositionPart)

                if neededCount < 0.001:
                    if isPart:
                        pos.setQuantity(0)
                    continue

                klPos = pos.source()
                if isPart:
                    # Если есть что-то неиспользуемое в исходной позиции засунем сюда
                    pos.setQuantity(pos.bquant + klPos.restCount())

                if pos.bquant < 0.001:
                    continue

                # определяемся с количеством, которое заберем у позиции
                if neededCount >= pos.bquant:
                    subCount = pos.bquant
                    if isPart and pos.isWholePos():
                        kitItem.positions[i] = klPos
                        pos = klPos
                        isPart = False
                else:
                    subCount = neededCount
                    if not isPart:
                        pos = klPos.createPart()
                        isPart = True
                    pos.setQuantity(subCount)
                neededCount = round(neededCount - subCount, 3)

                # объединение позиций
                if unitePositionsInKit and pos.isUniteAvailable():
                    posKey = klPos.uniqueKey()
                    if not collector:
                        collector = pos
                    elif collector.uniqueKey() != posKey:
                        self.applyDisc(collector, kitItem, occurrenceCount)
                        positionsForPriceDiscount.append(collector)
                        collector = pos
                    else:
                        collector.uniteWith(pos)
                        self.united.add(collector.source())
                else:
                    self.applyDisc(pos, kitItem, occurrenceCount)
                    positionsForPriceDiscount.append(pos)

            if unitePositionsInKit and collector:
                self.applyDisc(collector, kitItem, occurrenceCount)
                positionsForPriceDiscount.append(collector)

            # расчет цены набора после выполнения объединения/разбиения позиций, иначе цена может быть некорректной
            kitPrice = 0.0
            for positionForPriceDiscount in positionsForPriceDiscount:
                kitPrice += positionForPriceDiscount.sumn

            pureImpactsStart = len(self.pureImpacts)
            # применение воздействия Price
            for positionForPriceDiscount in positionsForPriceDiscount:
                self.applyDisc(positionForPriceDiscount, kitItem, occurrenceCount, kitPrice, True)

            pureImpactsEnd = len(self.pureImpacts)
            if (pureImpactsStart < pureImpactsEnd):
                # если цена набора не равна цене в воздействии Price, то корректируем скидку
                discountByPrice = 0.0
                for pureImpactsIndex in range(pureImpactsStart, pureImpactsEnd):
                    discountByPrice += self.pureImpacts[pureImpactsIndex].absoluteValue
                additionalDiscount = round(kitPrice - discountByPrice - kitItem.discRate.getValue() * occurrenceCount, 2)
                for pureImpactsIndex in range (pureImpactsStart, pureImpactsEnd):
                    changedAbsoluteValue = round(self.pureImpacts[pureImpactsIndex].absoluteValue + additionalDiscount, 2) 
                    if (changedAbsoluteValue > 0):
                        self.pureImpacts[pureImpactsIndex].absoluteValue = changedAbsoluteValue
                        break

    def createKit(self, kitCodes, posPerKitItem):
        """ Создание набора из кодов товаров @arg kitCodes на основе набора @arg posPerKitItem """
        kit = copy.deepcopy(posPerKitItem)
        for codeIndex in range(len(kitCodes)):
            kit[codeIndex].positions = self.allItems(posPerKitItem[codeIndex].positions, kitCodes[codeIndex])
        return kit

    def allItems(self, klPositions, code):
        """ Получение всех KlPosition из списка @arg klPositions с кодом @arg code """
        return [item for item in klPositions if item.getCode() == code]

    def getUniqueCodesByKit(self, posPerKitItem):
        """
        Получение списка списокв уникальных кодов для каждой части набора
        Результат: [['codeA', ..., 'codeB'], ['codeC', ..., 'codeD'], ..., ['codeY', ..., 'codeZ']]
        """
        return [list(set(klPosition.getCode() for klPosition in kitItem.positions)) for kitItem in posPerKitItem]

    def occ(self, posPerKitItem):
        """ Подготовка и расчёт количества вхождений набора в чек """
        # Сортируем позиции в частях набора
        for kitItem in posPerKitItem:
            kitItem.sort()

        # Делим между частями общие позиции
        self.distributeCommon(posPerKitItem)

        log.debug(u"Распределение позиций по частям набора - %s" % posPerKitItem)

        # Считаем количество вхождений набора
        return min(min([int(round(ki.count(), 3) / ki.kitItem.quantity) for ki in posPerKitItem]), self.kit.maxOccurances(self.check))
