# -*- coding: utf-8 -*-

import logging

from artixds.discountrate import DiscountRateType
from artixds.objectdiscount import PositionObject
from artixds.kitlogic.klpositionpart import KlPositionPart

log = logging.getLogger("KitLogic")


class KlKitItem(object):
    '''
    Часть набора
    Содержит позиции которые потециально входят в эту часть набора
    '''
    def __init__(self, kitItem, discRate, check):
        self.positions = []
        self.kitItem = kitItem
        self.discRate = discRate
        self.hasCommon = False
        self.check = check
        if self.discRate is not None:
            self.discRate.checkWrapper = check
        self.fract = kitItem.quantity % 1 > 0.0005  # У части кол-во весовое

    def append(self, pos):
        '''
        Добавление позиции
        '''
        if isinstance(pos, KlPositionPart):
            self.hasCommon = True
        self.positions.append(pos)

    def remove(self, pos):
        '''
        Удаление позиции
        '''
        self.positions.remove(pos)

    def uniqueCount(self, common):
        '''
        Сумма количества всех уникальных позиций
        '''
        result = 0
        for pos in self.positions:
            if isinstance(pos, KlPositionPart):
                if pos.parent in common:
                    continue
                result += pos.parent.bquant
                common.add(pos.parent)
            else:
                result += pos.bquant
        return result

    def count(self):
        '''
        сумма количества всех позиций
        '''
        result = 0
        for pos in self.positions:
            result += pos.bquant
        return result

    def distributeCommon(self, factor):
        '''
        Распределение общих позиций
        Забирает из общих позиций все что можно
        '''
        idealCount = self.kitItem.quantity * factor
        for pos in self.positions:
            idealCount = round(idealCount, 3)
            if idealCount < 0.001:
                break
            if isinstance(pos, KlPositionPart):
                subCount = min(pos.parent.restCount(), idealCount)
                pos.setQuantity(subCount)
                idealCount -= pos.bquant
            else:
                idealCount -= pos.bquant

    def sort(self):
        '''
        Сортировка позиций
         Если элемент набора имеет не суммовую скидку
           1. первыми идут позиции которые попадают только в эту часть набора
           2. первыми идут позиции с меньшей абсолютной скидкой на единицу товара
           3. одинаковые позиции идут вместе
         Если элемент набора НЕ имеет скидку или скидка суммовая
           1. первыми идут позиции которые попадают только в эту часть набора
           2. первыми идут позиции с большей ценой
           3. одинаковые позиции идут вместе

        Если в часть набора попали весовые и не весовые товары, то весовые будут последними.
        Вообще смешивать в одной части набора весовые и не весовые плохая идея,
        т к не весовые товары нельзя делить.
        '''

        if not self.positions:
            return

        sortItems = []

        # если есть и весовые и не весовые товары, то максимальный приоритет у не весовых
        prevFract = self.positions[0].source().fract
        for i in range(1, len(self.positions)):
            if prevFract != self.positions[i].source().fract:
                sortItems.append(lambda pos: pos.source().fract)
                log.warn(u"В части набора '%s' весовые и не весовые товары, приоритет у не весовых"
                        % self.kitItem.name)
                break

        # первые позиции которые попали только в эту часть набора
        sortItems.append(lambda pos: isinstance(pos, KlPositionPart))

        # сортировка по сумме скидки
        if self.discRate and self.discRate.getRateType() != DiscountRateType.SUM.value:
            # Модификатор позиции. Меняем при расчете скидки кол-во на 1 и суммы соответственно
            class PosMod(object):
                def __init__(self, pos):
                    pos.source().setQuantityForObject(self, 1.0)
            item = lambda pos: round(self.discRate.getAbsoluteValue(PositionObject(pos.source().pos.data, PosMod(pos)), self.check.checkDict), 2)
            sortItems.append(item)

        # сортировка по цене
        if self.discRate:
            sortItems.append(lambda pos: pos.source().pos.data['price'])
        else:
            sortItems.append(lambda pos: - pos.source().pos.data['price'])

        # сортировка по уникальному ключу
        sortItems.append(lambda pos: pos.source().uniqueKey())

        # сортировка по исходному номеру позиции
        sortItems.append(lambda pos: pos.source().posnum)

        def sortCriteria(pos):
            return tuple([item(pos) for item in sortItems])

        self.positions = sorted(self.positions, key=sortCriteria)

    def __repr__(self):
        return "KlKitItem(hasCommon=%s,positions=%s,quantity=%s)" % (self.hasCommon, self.positions, self.kitItem.quantity)
