# -*- coding: utf-8 -*-
import copy

from artixds.kitlogic.klpositionpart import KlPositionPart


class KlPosition(object):
    """ Позиция в части набора
    parts - содержит части позиции
    """
    def __init__(self, pos):
        self.pos = pos
        self.kitItemNum = None
        self.parts = []
        self.fract = pos.data['tmcunit']['fractional']

        # Эти поля называются как в позиции, это сделано специально
        self.bquant = pos.data['bquant'] if self.fract else int(pos.data['bquant'])
        self.posnum = pos.data['posnum']
        self.sumn = pos.data['sumn']
        self.sumb = pos.data['sumb']
        self.partOfKit = False

        # Позиции которые были объединены с этой, хранит {<номер позиции>: <кол-во>}
        self.unitedPositions = {}

    def getCode(self):
        return self.pos.data['code']

    def isUniteAvailable(self):
        return self.pos.data['uniteavailable'] if 'uniteavailable' in self.pos.data.data else True

    def source(self):
        """ Возвращает исходную позицию """
        return self

    def setQuantityForObject(self, object, bquant):
        """
        Общий метод для изменения количества позиции или части позиции

        Если не было скидок на позицию, то пересчитываем sumb от цены, иначе пропорционально
        Если пересчет пропорциональный, то есть вероятность потерять точность sumb
        """
        object.bquant = round(bquant, 3) if self.fract else int(bquant)
        data = self.pos.data
        object.sumn = data['price'] * object.bquant
        if abs(data['sumn'] - data['sumb']) < 0.005:
            object.sumb = object.sumn
        else:
            object.sumb = data['sumb'] * object.bquant / data['bquant']

    def setQuantity(self, bquant):
        """ Изменение количества с пересчетом сумм """
        self.setQuantityForObject(self, bquant)

    def uniteWith(self, pos):
        """
        Объединение позиции pos с этой
        pos - позиция или часть позиции которая добавляется к текущей
        """
        self.setQuantity(self.bquant + pos.bquant)
        srcPos = pos.source()
        self.unitedPositions[srcPos.posnum] = pos.bquant
        if isinstance(pos, KlPositionPart):
            srcPos.setQuantity(srcPos.bquant - pos.bquant)
        pos.setQuantity(0)

    def hasUnited(self):
        """ Возвращает True если была объединена с другими позициями """
        return bool(len(self.unitedPositions))

    def uniqueKey(self):
        """
        Уникальный ключ позиции
        используется для объединения позиций
        """
        return self.pos.data['bcode'], self.pos.data['code'], self.pos.data['price']

    def usedTimes(self):
        """ Количество использований в разных частях набора """
        if self.parts:
            return len(self.parts)
        elif self.kitItemNum is not None:
            return 1
        return 0

    def createPart(self):
        """ Создает часть позиции """
        part = KlPositionPart(self)
        self.parts.append(part)
        return part

    def restCount(self):
        """ Остатки. Количество которое не попало в части позиции """
        return self.bquant - sum([i.bquant for i in self.parts]) - sum(self.unitedPositions.values())

    def complete(self, posNum):
        """
        Удаляет пустые части позиции
        Раскидывает новые номера, если не был установлен ранее
        Если есть неиспользованное количество создает часть с остатками
        posNum - функция которая возвращает номер позиции
        """
        partsCount = 0

        newParts = []
        for part in self.parts:
            if part.isWholePos():
                break
            if part.bquant < 0.001:
                continue
            partsCount += part.bquant
            if not part.posnum:
                part.posnum = posNum()
            newParts.append(part)
        self.parts = newParts

        rest = self.bquant - partsCount
        if self.parts and rest > 0.0001:
            part = self.createPart()
            part.setQuantity(rest)
            part.posnum = posNum()

    def __repr__(self):
        return "KlPosition(code=%s,bquant=%s,posnum=%s)" % (repr(self.pos.data['code']), self.bquant, self.posnum)
