# -*- coding: utf-8 -*-
'''
Created on 31.08.2010

@author: gsu
'''

from artixds.kitentry import KitEntry, KitEntryItem, KitEntryPosition
from artixds.impact import Impact, ImpactType, ImpactObjectPosition
from artixds.branchdetail import BranchDetail
from artixds.objectdiscount import PositionObjectModifiers, PositionObject, KitType
import logging
from artixds.impact import ImpactMode

log = logging.getLogger("Roster")


class RosterItemObject(object):
    def __init__(self, code, modifier, rosterItem, catalogs=None):
        self.code = code
        self.modifier = modifier
        self.rosterItem = rosterItem
        self.catalogs = catalogs

    def __repr__(self):
        # временная заглушка, т.к. нужно разобраться с кириллическими символами в кодах товаров
        return "RosterItem(positionNumber = %d, code = ???, quantity = %.3f, sumn = %.2f, sumb = %.2f, %s)" % \
               (self.positionNumber, self.quantity, self.sumn, self.sumb,
                "part of kit" if self.modifier.partOfKit else "vacant item")

    positionNumber = property(lambda self: self.modifier.posnum)
    quantity = property(lambda self: self.modifier.bquant)
    sumn = property(lambda self: self.modifier.sumn)
    sumb = property(lambda self: self.modifier.sumb)
    kitEntryPosition = property(lambda self: self.modifier.kitEntryPosition)

    def setImpact(self, impact):
        self.modifier.impact = impact

    impact = property(lambda self: self.modifier.impact, setImpact)

    def setBorrow(self, value):
        self.rosterItem.borrowed = value

    borrowed = property(lambda self: self.rosterItem.borrowed, setBorrow)

    def isPartOfKit(self):
        return self.modifier.partOfKit

    def subtractKitEntryPosition(self, kitEntryPosition):
        self.rosterItem.subtractKitEntryPosition(kitEntryPosition)


class RosterItem(object):
    '''
    Позиция хранит историю своих изменений в виде списка.
    Каждый элемент списка может быть активным или нет (позиция, к которой были применены структурные воздействия).
    Часть из них входит в комплекты, другие нет.
    '''

    def __init__(self, data, roster):
        self.data = data
        self.roster = roster
        self.modifiedItems = [
            PositionObjectModifiers(data["posnum"], data["partOfKit"] if 'partOfKit' in data else False, data["cquant"], \
                                    data["bquant"], data["sumi"], data["sumn"], data["sumb"], data["sume"],
                                    data["packingmode"])]
        self.borrowed = 0

    @classmethod
    def rosterItemWithData(cls, data, roster):
        return RosterItem(data, roster)

    def __repr__(self):
        return "RosterItem(positionNumber = %d, code = %s, quantity = %.3f, sumn = %.2f, sumb = %.2f)" % \
               (self.data['posnum'], repr(self.data['code']), self.data['cquant'], self.data['sumn'], self.data['sumb'])

    def __iter__(self):
        '''
        Итерация по конечным позицям
        '''
        for item in self.modifiedItems:
            if item.active:
                yield RosterItemObject(self.data['code'], item, self, self.data['tmc']['catalogs'])

    def vacantIter(self):
        '''
        Позиции, не занятые в комплектах
        '''
        for item in self.modifiedItems:
            if not item.partOfKit and item.active and not item.packingMode:
                yield RosterItemObject(self.data['code'], item, self, self.data['tmc']['catalogs'])

    def kitIter(self):
        '''
        Позиции, занятые в комплектах
        '''
        for item in self.modifiedItems:
            if item.kitEntryPosition and item.active:
                yield RosterItemObject(self.data['code'], item, self, self.data['tmc']['catalogs'])

    def root(self):
        '''
        Возвращает исходную позицию, от которой произошла эта
        '''
        return self

    def subtractKitEntryPosition(self, kitEntryPosition):
        pos = list(filter(lambda x: x.active and not x.kitEntryPosition, self.modifiedItems))[0]
        pos.active = False
        self.modifiedItems.append(
            self.createNewItem(pos, min(kitEntryPosition.quantity, pos.bquant), True, kitEntryPosition))
        if pos.bquant - kitEntryPosition.quantity > 0:
            self.modifiedItems.append(
                self.createNewItem(pos, pos.bquant - kitEntryPosition.quantity, pos.partOfKit, None))

        for item in self.modifiedItems:
            log.debug("%s, positionNumber = %d, code = %s, quantity = %.3f, sumn = %.2f, sumb = %.2f, %s" % (
            "+" if item.active else "-", \
            item.posnum, self.data['code'], item.bquant, item.sumn, item.sumb,
            "part of kit" if item.partOfKit else "vacant item"))

    def createNewItem(self, item, quantity, partOfKit, kitEntry):
        #        posnum = self.roster.nextPositionNumber()
        posnum = item.posnum if (item.bquant - quantity) < 0.0001 else self.roster.nextPositionNumber()
        newItem = PositionObjectModifiers(posnum, partOfKit, item.cquant * quantity / item.bquant, quantity, \
                                          item.sumi * quantity / item.bquant, item.sumn * quantity / item.bquant, \
                                          item.sumb * quantity / item.bquant, item.sume * quantity / item.bquant,
                                          item.packingMode)
        newItem.kitEntryPosition = kitEntry
        return newItem


class Roster(object):
    '''
    Упрощённое представление чека. Содержит только коды товаров с количеством.
    '''

    def __init__(self, check):
        self.check = check
        self.items = [RosterItem.rosterItemWithData(it.data, self) for it in check.getGoodsList()]
        self._nextPositionNumber = self.check.getNextPositionNumber()

    def __repr__(self):
        return u"Roster(items=%s)" % self.items

    def __doIter(self, iterator):
        '''
        Общая итерация по позициям
        '''
        for item in self.items:
            for subitem in iterator(item):
                yield subitem

    def __iter__(self):
        '''
        Итерация по позициям чека
        '''
        for it in self.__doIter(RosterItem.__iter__):
            yield it

    def vacantIter(self):
        '''
        Итерация по не занятым в комплектах позициям чека
        '''
        for it in self.__doIter(RosterItem.vacantIter):
            yield it

    def kitIter(self):
        '''
        Итерация по занятым в комплектах позициям чека
        '''
        for it in self.__doIter(RosterItem.kitIter):
            yield it

    def __kitSort(self, kit):
        '''
        Сортировка позиций набора. 
        Возвращает отсортированный набор    
        '''
        convolution = {}
        for item in self.vacantIter():
            for catalog in list(map(lambda x: x["catalogid"], item.catalogs)):
                if catalog not in convolution:
                    convolution[catalog] = 0
                convolution[catalog] += item.quantity
        kit.items.sort(key=lambda x: (convolution[x.codes[0]] if x.codes[0] in convolution else 0) - x.quantity)
        return kit

    def __findKitItem(self, kitItem, kitType):
        '''
        Ищет KitItem в чеке  и возвращает вхожденние в виде KitEntryItem
        kitType - тип набора
        '''
        result = KitEntryItem(kitItem=kitItem)
        quantityLeft = kitItem.quantity
        for rosterItem in self.vacantIter():
            catalogs = set(map(lambda x: x["catalogid"], rosterItem.catalogs))
            if (kitType == KitType.CATALOGS.value and set(kitItem.codes) & catalogs) or (
                    kitType == KitType.CODES.value and rosterItem.code in kitItem.codes):
                if quantityLeft > 0.0 and (rosterItem.quantity - rosterItem.borrowed) > 0.0:
                    quantityAdd = min(rosterItem.quantity - rosterItem.borrowed, quantityLeft)
                    result.addPosition(
                        KitEntryPosition(rosterItem=rosterItem, quantity=quantityAdd, kitEntryItem=result))
                    quantityLeft -= quantityAdd
                    rosterItem.borrowed += quantityAdd
        if result.quantity() == kitItem.quantity:
            return result
        return None

    def findKit(self, kit):
        '''
        Ищет KitObject в чеке и возвращает вхождение в виде KitEntry
        kit - KitObject
        возвращает KitEntry или None
        '''

        if not kit.kittype in [KitType.CODES.value, KitType.CATALOGS.value]:
            kit.kittype = KitType.CODES.value

        for rosterItem in self.vacantIter():
            rosterItem.borrowed = 0.0
        result = KitEntry()
        # если набор задан каталогами,
        # то сортируем его 
        if (kit.kittype == KitType.CATALOGS.value):
            kit = self.__kitSort(kit)
        for kitItem in kit:
            kitEntryItem = self.__findKitItem(kitItem, kit.kittype)
            if kitEntryItem:
                result.addItem(kitEntryItem)
            else:
                return None
        return result if result.items else None

    def subtractKitEntry(self, kitEntry):
        '''
        Выделяет KitEntry (вхождение комплекта) из чека
        '''
        for position in kitEntry.positions():
            position.rosterItem.subtractKitEntryPosition(position)

    def structureImpacts(self):
        '''
        Возвращает конечные структурные воздействия на чек
        '''
        impacts = []
        for item in self.items:
            branches = list(filter(lambda x: x.rosterItem.data['posnum'] != x.positionNumber, item))
            if not branches:
                continue
            impacts.append(Impact.branchImpact(ImpactObjectPosition(item.data['posnum'], item.data['code']), \
                                               [BranchDetail(i.positionNumber, i.quantity, i.sumn, i.sumb,
                                                             i.isPartOfKit()) for i in branches]))
        return impacts

    def partsOfKit(self):
        return [i.positionNumber for i in self.kitIter()]

    def pureImpacts(self):
        impacts = []
        for item in self:
            if item.impact:
                impacts.append(item.impact)
        return impacts

    def nextPositionNumber(self):
        '''
        Генератор номера позиции
        '''
        posnum = self._nextPositionNumber
        self._nextPositionNumber += 1
        return posnum

    def addItem(self, item):
        '''
        Кладёт позицию в чек
        '''
        self.items.append(item)
        self.lastPositionNumber = max([self.lastPositionNumber, item.positionNumber])

    def applyRate(self, kitDiscountRate, campaign=None, discount=None, discountContext=None):
        log.debug('applying rates')

        # play with roster items
        # ПОЗИЦИИ ФИНАЛЬНОГО ЧЕКА. КАЖДАЯ ИЗ НИХ ВХОДИТ В ОДНУ ИЛИ НИ ОДНОЙ KitEntryItems
        # ЧТОБЫ ПОНЯТЬ В КАКУЮ, СДЕЛАЕМ ТАК.
        if log.isEnabledFor(logging.DEBUG):
            log.debug('1: %s' % [i for i in self.kitIter()])
            log.debug('2: %s' % [i for i in self])
            log.debug('3: %s' % [i for i in self.vacantIter()])
            rosterItems = [i for i in self.kitIter()]
            log.debug('kitEntryItems: %s' % rosterItems)

            l = [i for i in self.kitIter()]
            log.debug('1: %s' % l)

        kitEntryItems = {}
        # выберем все присутствующие в чеке РАЗЛИЧНЫЕ kitEntryItems (вхождения элементов комплекта)
        for item in self.kitIter():
            kitEntryItems[item.kitEntryPosition.kitEntryItem] = 1
        log.debug('kitEntryItems: %s', kitEntryItems)
        for kitEntryItem, _ in kitEntryItems.items():
            # выберем позиции чека, входящие в заданный элемент комплекта
            rosterItems = kitEntryItem.rosterItems()
            log.debug('kit discount rate: %s', kitDiscountRate)
            discountRate = kitDiscountRate.discountRateForItem(kitEntryItem.kitItem)
            log.debug('discount rate: %s', discountRate)

            if discountRate:
                rateValues = discountRate.getAbsoluteValue(kitEntryItem, self.check)
                for x, value in zip(rosterItems, rateValues):
                    impact = Impact(type=ImpactType.PURE.value,
                                    campaign=campaign,
                                    discount=discount,
                                    absoluteValue=value,
                                    impactObject=ImpactObjectPosition(positionNumber=x.positionNumber, code=x.code),
                                    qty=x.quantity,
                                    mode=ImpactMode.BYDISCKIT.value,
                                    discountRate=discountRate)
                    if discountContext is not None:
                        impact.cards = discountContext.cards
                        impact.counters = discountContext.counters
                        impact.coupons = discountContext.coupons
                    x.setImpact(impact)


class RosterFactory(object):
    @staticmethod
    def getRosterByCheck(check):
        result = Roster(check)
        return result
