# -*- coding: utf-8 -*-
'''
Created on 14.02.2011

@author: gsu
'''

from legacydiscount.disccardaccumulate import DiscCardAccumulate
from legacydiscount.checktotals import DiscountObjectForCheck, DiscountObjectForPosition, DiscountCardObject
from legacydiscount.discauto import AutomaticDiscount
from legacydiscount.disccard import DiscCard
from legacydiscount.connection.connection import Session
from sqlalchemy import Table, MetaData, String, Numeric, Column
from sqlalchemy.sql import or_, func
from legacydiscount.compare import coalesce
from legacydiscount.kit import Kit
from legacydiscount.discountsfixed import FixedDiscount
import datetime
import logging
from legacydiscount.log_decorator import log_after
from enum import Enum
from legacydiscount.discountrate import ZeroDiscountRate
from legacydiscount.sprices import Sprices
from legacydiscount.uniq import uniq
from config import Config

session = Session()

logger = logging.getLogger('su.artix.dm')


class DiscountMode(Enum):
    AUTOMATIC = 'automatic'
    BYCARD = 'bycard'
    MANUAL = 'manual'
    FIXED = 'fixed'
    BYDISCKIT = 'bydisckit'
    BYPAYMENT = 'bypayment'


class AccumulativeDiscountManager(object):

    def __repr__(self):
        return u'AccumulativeDiscountManager'

    @log_after
    def accumulative(self, card):
        logger.debug('trying accumulate discounts for card : %s (%s)' % (card.cardnumber, card.name))
        logger.debug('accumulative discounts scheme        : %s' % (card.acdiscasheme))
        logger.debug('cardsum                              : %s' % (card.cardsum))
        logger.debug('rcount                               : %s' % (card.rcount))

        DCA = DiscCardAccumulate
        query = session.query(DCA)
        query = query.filter(DCA.discscheme == card.acdiscasheme)
        query = query.filter(or_(func.coalesce(DCA.isbegsum, 0) == 0, DCA.begsum <= coalesce(card.cardsum, 0)))
        query = query.filter(or_(func.coalesce(DCA.isendsum, 0) == 0, DCA.endsum >= coalesce(card.cardsum, 0)))
        query = query.filter(or_(func.coalesce(DCA.isbegrcount, 0) == 0, DCA.begrcount <= coalesce(card.rcount, 0)))
        query = query.filter(or_(func.coalesce(DCA.isendrcount, 0) == 0, DCA.endrcount >= coalesce(card.rcount, 0)))
        query.order_by(DCA.disccode)

        res = query.all()

        logger.debug('%d discounts matched' % (len(res)))

        return [res[0]] if res else []

    @log_after
    def discountToCheck(self, check, cardrecord):
        logger.debug('trying accumulative discount for check: %s' % (check))
        disccard = cardrecord.disccard if cardrecord else None
        rates = self.accumulative(disccard)
        return [Discount(check, check, check, rate, cardrecord, DiscountMode.BYCARD.value) for rate in rates]


class AutomaticDiscountManager(object):
    defaultItemPriceType = Config().getInt('Check', 'defaultItemPriceType', 0)
    defaultCheckPriceType = Config().getInt('Check', 'defaultCheckPriceType', 0)
    discountSchemes = {}
    discountSchemesHeaders = {}
    catalogs = {}

    def __init__(self):
        logger.debug('defaultItemPriceType = %d' % self.defaultItemPriceType)
        logger.debug('defaultCheckPriceType = %d' % self.defaultCheckPriceType)

    def __repr__(self):
        return u'AutomaticDiscountManager'

    @log_after
    def loadSchemesFromDatabase(self, schemes, clean=True):
        logger.debug('load a data for schemes: %s' % schemes)
        if clean == True:
            self.discountSchemesHeaders = {}
            self.discountSchemes = {}
        DA = AutomaticDiscount
        query = session.query(DA)
        query = query.filter(DA.shemecode.in_(schemes))
        query = query.order_by(DA.shemecode, DA.disccode)
        result = query.all()
        for rec in result:
            if not str(rec.shemecode) in self.discountSchemes:
                self.discountSchemes[str(rec.shemecode)] = []
            self.discountSchemes[str(rec.shemecode)].append(rec)

        S = Sprices
        query = session.query(S)
        query = query.filter(S.code.in_(schemes))
        result = query.all()
        for rec in result:
            self.discountSchemesHeaders[str(rec.code)] = rec

    @log_after
    def auto(self, object, check, schemeCode, card=None):
        logger.debug('trying automatic discounts')
        cardgroupcode = card.groupcode if card and card.groupcode else 0
        birthday = card.birthday if card and card.birthday else None
        if logger.isEnabledFor(logging.DEBUG):
            if card:
                logger.debug('card			: %s (%s)' % (card.cardnumber, card.name))
            logger.debug('scheme code	 : %s' % (schemeCode))
            logger.debug('card group code : %s' % (cardgroupcode))
            logger.debug('bithday		 : %s' % (birthday))
            logger.debug('time			: %s' % (object.time))
            logger.debug('date			: %s' % (object.date))
            logger.debug('weekday		 : %s' % (object.weekday))
            logger.debug('cquant		  : %s' % (object.cquant))
            logger.debug('sumn			: %s' % (object.sumn))

        res = []
        if (str(schemeCode) not in self.discountSchemesHeaders) or (str(schemeCode) not in self.discountSchemes):
            return res
        if (not card) and (self.discountSchemesHeaders[str(schemeCode)].type == 1):
            return res

        catalogKeys = self.catalogs.keys()
        for disc in self.discountSchemes[str(schemeCode)]:
            if disc.catalogid:
                id = str(disc.catalogid)
                if id not in catalogKeys:
                    continue
                if disc.catalogquantbeg and (disc.catalogquantbeg > self.catalogs[id]['quant']):
                    continue
                if disc.catalogquantend and (disc.catalogquantend < self.catalogs[id]['quant']):
                    continue
                if disc.catalogsumbeg and (disc.catalogsumbeg > self.catalogs[id]['sum']):
                    continue
                if disc.catalogsumend and (disc.catalogsumend < self.catalogs[id]['sum']):
                    continue
            if cardgroupcode != (disc.cardgroupcode if disc.cardgroupcode else 0):
                continue
            if disc.birthdaymode and (disc.birthdaymode == 1) and (
                    birthday == None or ((object.date.day != birthday.day) or (object.date.month != birthday.month))):
                continue
            if (disc.date_beg and (object.date < disc.date_beg)) or (disc.date_end and (object.date > disc.date_end)):
                continue
            if (disc.time_beg and (object.time < disc.time_beg)) or (disc.time_end and (object.time > disc.time_end)):
                continue
            if (disc.days_beg and (disc.days_beg != 0) and (object.weekday < disc.days_beg)) or (
                    disc.days_end and (disc.days_end != 0) and (object.weekday > disc.days_end)):
                continue
            if (disc.quant_beg and (object.cquant < disc.quant_beg)) or (
                    disc.quant_end and (object.cquant > disc.quant_end)):
                continue
            if disc.sum_beg and (
                    disc.sum_beg > (object.sumn if (disc.rsummode == None) or (disc.rsummode == 0) else check.sumn)):
                continue
            if disc.sum_end and (
                    disc.sum_end < (object.sumn if (disc.rsummode == None) or (disc.rsummode == 0) else check.sumn)):
                continue

            res.append(disc)
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug('discount with code %s' % disc.disccode)
                logger.debug('	mode is %s' % disc.discmode)
                logger.debug('	type is %s' % disc.disctype)
                logger.debug('	val is %s' % disc.discval)
            break
        return res

    @log_after
    def discountToItem(self, object, check, cardrecord=None):
        logger.debug('trying automatic discount for goods item: %s' % (object))
        disccard = cardrecord.disccard if cardrecord else None
        schemeCode = object.pricetype if object.pricetype else self.defaultItemPriceType
        cardScheme = disccard.discasheme if disccard else None
        if cardScheme and (disccard.allpositions if disccard.allpositions else 0) == 0 and cardScheme != schemeCode:
            return []
        if cardScheme and (disccard.allpositions if disccard.allpositions else 0) == 1:
            schemeCode = cardScheme
        rates = self.auto(object, check, schemeCode, disccard)
        return [Discount(check, object, object, rate, cardrecord,
                         DiscountMode.AUTOMATIC.value if cardrecord == None else DiscountMode.BYCARD.value) for rate in rates]

    @log_after
    def discountToCheck(self, check, cardrecord=None):
        logger.debug('trying automatic discount for check: %s' % (check))
        disccard = cardrecord.disccard if cardrecord else None
        schemeCode = check.pricetype if check.pricetype else self.defaultCheckPriceType
        if disccard and disccard.discasheme: schemeCode = disccard.discasheme
        rates = self.auto(check, check, schemeCode, disccard)
        return [Discount(check, check, check, rate, cardrecord,
                         DiscountMode.AUTOMATIC.value if cardrecord == None else DiscountMode.BYCARD.value) for rate in rates]


class FixedDiscountManager(object):

    @log_after
    def fixed(self, object, cardnumber):
        FD = FixedDiscount
        query = session.query(FD)
        query = query.filter(FD.cardpref != u'')

        if isinstance(object, DiscountObjectForCheck):
            query = query.filter(FD.receiptmode == 1)
        else:
            query = query.filter(FD.positionmode == 1)

        sumb = float(object.sumb)
        query = query.filter(func.coalesce(FD.cardsum_beg, sumb) <= sumb)
        query = query.filter(func.coalesce(FD.cardsum_end, sumb) >= sumb)
        query = query.order_by(FD.disccode)

        res = [item for item in query.all() if cardnumber.startswith(item.cardpref)]

        return [res[0]] if res else []

    @log_after
    def discountByCard(self, check, object, card):
        return [Discount(check, object, object, rate, card, DiscountMode.BYCARD.value) for rate in
                self.fixed(object, card.card.number)]


class Discount(object):
    check = None
    object = None
    rate = None
    card = None
    innerObject = None

    def __init__(self, check, object, innerObject, rate, card, mode):
        self.check = check
        self.object = object
        self.rate = rate
        self.card = card
        self.discountMode = mode
        self.innerObject = innerObject
        self.absoluteValue = self.rate.appliedValue(self.innerObject)
        self.deniedPositions = []

    def toDict(self):
        return {
            'discountRate': {'type': self.rate.type, 'value': float(self.rate.value)},
            'discountMode': self.discountMode,
            'absoluteValue': float(self.absoluteValue),
            'positionNumber': self.innerObject.posnum,
            'campaign': {'id': self.rate.scheme.code, 'name': self.rate.scheme.name},
            'discount': {'id': self.rate.disccode, 'name': self.rate.name, 'value': float(self.rate.value)},
            'cards': [self.card.dump()] if self.card else [],
            'deniedPositions': self.deniedPositions
        }

    def __repr__(self):
        if not self.card:
            return u'Discount (check = %s, object = %s, innerObject = %s, rate = %s, mode = %s)' % (
            self.check, self.object, self.innerObject, self.rate, self.discountMode)
        else:
            return u'Discount (check = %s, object = %s, innerObject = %s, rate = %s, mode = %s, card = %s)' % (
            self.check, self.object, self.innerObject, self.rate, self.discountMode, self.card)


class DiscountCollection(object):
    class Over:
        def __call__(self, all, next):
            return [next]

    class Deny:
        def __call__(self, all, next):
            return [(all + [next])[0]]

    class Min:
        def __call__(self, all, next):
            return [min(all + [next], key=lambda x: x.absoluteValue)]

    class Max:
        def __call__(self, all, next):
            return [max(all + [next], key=lambda x: x.absoluteValue)]

    class First:
        def __call__(self, all, next):
            return [(all + [next])[0]]

    class Last:
        def __call__(self, all, next):
            return [next]

    class All:
        def __call__(self, all, next):
            return all + [next]

    class Add:
        def __call__(self, all, next):
            return all + [next]

    discounts = []

    def __init__(self, discounts=None):
        self.discMode = Config().getString('Discounts', 'discMode', 'over').lower()
        self.discPositionMode = Config().getString('Discounts', 'discPositionMode', 'deny').lower()
        self.discRelation = Config().getString('Discounts', 'pos2CheckDiscRel', 'deny').lower()

        logger.debug(u"discount collection with discMode = '%s', discPositionMode = '%s' created." % (
        self.discMode, self.discPositionMode))
        if self.discMode == "over":
            self.checkMod = self.Over()
        elif self.discMode == "deny":
            self.checkMod = self.Deny()
        elif self.discMode == "all":
            self.checkMod = self.All()
        else:
            self.checkMod = self.Over()

        if self.discPositionMode == "over":
            self.posMod = self.Over()
        elif self.discPositionMode == "deny":
            self.posMod = self.Deny()
        elif self.discPositionMode == "all":
            self.posMod = self.All()
        elif self.discPositionMode == "max":
            self.posMod = self.Max()
        else:
            self.posMod = self.Deny()

        self.discounts = []
        if discounts:
            self.extend(discounts)

    def __iter__(self):
        for discount in self.discounts:
            yield discount

    def __repr__(self):
        return 'DiscountCollection of %d items' % (len(self.discounts))

    def _appendDiscount(self, discount):
        if isinstance(discount.innerObject, DiscountObjectForCheck) and self.discRelation == u'deny':
            positions = uniq(
                [d.innerObject.posnum for d in self.discounts if isinstance(d.innerObject, DiscountObjectForPosition)])
            discount.deniedPositions = positions

        all = [d for d in self.discounts if d.innerObject == discount.innerObject]
        for d in all:
            self.discounts.remove(d)
        all = self.checkMod(all, discount) if isinstance(discount.innerObject, DiscountObjectForCheck) else self.posMod(
            all, discount)
        self.discounts.extend(all)

    def appendDiscount(self, check, object, innerObject, rate):
        discount = Discount(check, object, innerObject, rate)
        self._appendDiscount(discount)

    @log_after
    def extend(self, discounts):
        for disc in discounts:
            self._appendDiscount(disc)

    def __bool__(self):
        return bool(self.discounts)

    def __nonzero__(self):
        return bool(self.discounts)

class DiscountManager(object):

    def __init__(self, totals):
        self.totals = totals
        self.adm = AutomaticDiscountManager()

    def __repr__(self):
        return u'DiscountManager'

    @log_after
    def discount(self, check):
        res = DiscountCollection()
        self.adm.loadSchemesFromDatabase(self.totals.getPricetypes(), True)
        self.adm.catalogs = self.totals.getCatalogs()
        checkObject = DiscountObjectForCheck(check)
        for object in self.totals.goodsItemsIter():
            res.extend(self.adm.discountToItem(object, checkObject))
        res.extend(self.adm.discountToCheck(checkObject))
        return res

    @log_after
    def discountByCard(self, check, disccard):
        card = DiscountCardObject(disccard)
        card.disccard = DiscCard.select(card.card.number)
        checkObject = DiscountObjectForCheck(check)

        if not card.disccard:
            logger.debug(u"no cards found with number '%s'" % card.card.number)

        res = DiscountCollection()

        if card.disccard:
            if card.disccard.acdiscasheme:
                res.extend(AccumulativeDiscountManager().discountToCheck(checkObject, card))

            if card.disccard.discasheme and (str(card.disccard.discasheme) not in self.adm.discountSchemesHeaders):
                self.adm.loadSchemesFromDatabase([card.disccard.discasheme], False)
            if not res:
                if card.disccard.positionmode or not card.disccard.discasheme:
                    # here we go with loop for items
                    for object in self.totals.goodsItemsIter():
                        res.extend(self.adm.discountToItem(object, checkObject, card))
                elif card.disccard.discasheme:
                    res.extend(self.adm.discountToCheck(checkObject, card))

            return res

        dm = FixedDiscountManager()
        # here we go with loop for items
        for object in self.totals.goodsItemsIter():
            res.extend(dm.discountByCard(checkObject, object, card))
        res.extend(dm.discountByCard(checkObject, checkObject, card))

        return res


class KitManager(object):

    def __init__(self):
        metadata = MetaData()
        self.goodsTable = Table('tmpshortgoodslist', metadata,
                                Column('code', String(100), primary_key=True),
                                Column('quantity', Numeric(precision=15, scale=3)),
                                mysql_engine='Memory',
                                mysql_charset='utf8')
        # self.goodsTable.create(bind = session.get_bind(None), checkfirst = True)

    def __repr__(self):
        return u'KitManager'

    @log_after
    def extractKits(self, check):
        # Поиск и выделение наборов составленных из каталогов
        for kit in Kit.selectActiveForCatalogs(datetime.datetime.now(), check.catalogsItemsShort(), self.goodsTable):
            check.distributionKit(kit)

        # Поиск и выделение наборов составленных из кодов товаров
        for kit in Kit.selectActiveForCodes(datetime.datetime.now(), check.goodsItemsShort(), self.goodsTable):
            times = check.times(kit)
            if times > 0:
                logger.debug('kit with code %s found %s times' % (kit.kitid, times))
                check.subtract(kit, times)

    @log_after
    def discounts(self, check):
        res = DiscountCollection()
        for kitEntry in check.kitEntries:
            if not kitEntry.kitgoods.kitdiscs:
                res.extend(
                    [Discount(check, kitEntry.kit, item, ZeroDiscountRate(), None, DiscountMode.BYDISCKIT.value) for item in
                     kitEntry.items])
            else:
                for rate in kitEntry.kitgoods.kitdiscs:
                    rate.setTimes(kitEntry.times)
                    rate.appliedToObjects(kitEntry.items)
                    res.extend([Discount(check, kitEntry.kit, item, rate, None,
                                         DiscountMode.BYPAYMENT.value if rate.impactmode == 2 else DiscountMode.BYDISCKIT.value) for
                                item in kitEntry.items])
        return check.branchImpacts(), res
