# -*- coding: utf-8 -*-

from simplediscount.uniq import uniq
from simplediscount.discountobject import DiscountObjectMixin
from decimal import Decimal
from datetime import date, time
from simplediscount.compare import coalesce
from simplediscount.kitentry import KitEntry
import logging
logger = logging.getLogger('su.artix.dm')


class DiscountObjectForPosition(DiscountObjectMixin):
    _Zero = Decimal('0.00')
    
    def __init__(self, position, modifiers = None):
        self._position = position
        self._modifiers = modifiers
    
    def __eq__(self, other):
        return self.posnum == other.posnum

    code = property(lambda self: self._position.code)
    catalogs = property(lambda self: self._position.getTmcObject().getCatalogsList())
    price = property(lambda self: Decimal("%0.02f" % self._position.price))
    posnum = property(lambda self: self._modifiers.posnum if self._modifiers else self._position.posnum)
    partOfKit = property(lambda self: self._modifiers.partOfKit if self._modifiers else False)
    bquant = property(lambda self: Decimal("%0.03f" % (self._modifiers.bquant if self._modifiers else self._position.bquant)))
    cquant = property(lambda self: Decimal("%0.03f" % (self._modifiers.cquant if self._modifiers else self._position.cquant)))
    sumn = property(lambda self: Decimal("%0.02f" % (self._modifiers.sumn if self._modifiers else self._position.sumn)))
    sumb = property(lambda self: Decimal("%0.02f" % (self._modifiers.sumb if self._modifiers else self._position.sumb)))
    pricetype = property(lambda self: self._position.pricetype if self._position.pricetype != 0 else None)
    time = property(lambda self: time(self._position.item_datetime.time().hour(), self._position.item_datetime.time().minute(), self._position.item_datetime.time().second()))
    date = property(lambda self: date(self._position.item_datetime.date().year(), self._position.item_datetime.date().month(), self._position.item_datetime.date().day()))
    weekday = property(lambda self: self._position.item_datetime.date().dayOfWeek())
    kitgoods = property(lambda self: self._modifiers.kitgoods if self._modifiers else None)
    packingMode = property(lambda self: self._position.packingMode)

    def indexPrice(self, index):
        obj = self._position.getTmcObject().getIndexPriceObject(index)
        return Decimal("%0.02f" % (obj.price if obj else self._position.price))


class DiscountObjectForCheck(DiscountObjectMixin):
    _Zero = Decimal('0.000')
    
    def __init__(self, check):
        self._check = check

    def __eq__(self, other):
        return self.posnum == other.posnum

    posnum = property(lambda self: -1)
    partOfKit = property(lambda self: False)
    bquant = property(lambda self: Decimal("%0.03f" % (self._check.quantity)))
    cquant = property(lambda self: Decimal("%0.03f" % (self._check.quantity)))
    sumn = property(lambda self: Decimal("%0.02f" % self._check.sum1))
    sumb = property(lambda self: Decimal("%0.02f" % self._check.sumb))
    pricetype = property(lambda self: self._check.priceType if self._check.priceType != 0 else None)
    time = property(lambda self: time(self._check.dateTimeBeg.time().hour(), self._check.dateTimeBeg.time().minute(), self._check.dateTimeBeg.time().second()))
    date = property(lambda self: date(self._check.dateTimeBeg.date().year(), self._check.dateTimeBeg.date().month(), self._check.dateTimeBeg.date().day()))
    weekday = property(lambda self: self._check.dateTimeBeg.date().dayOfWeek())

    def indexPrice(self, index):
        return self._Zero
    
    def getCardsList(self):
        return self._check.getCardsList()


class DiscountCardObject:
    def __init__(self, card):
        self._card = card

    card = property(lambda self: self._card.getCardObject())
    
    def dump(self):
        return self._card.getDump()    


class DiscountKitModifiers:
    def __init__(self, posnum, partOfKit, cquant, sumn, bquant=None, sumb=None, kitgoods=None):
        self.posnum = posnum
        self.cquant = cquant
        self.bquant = bquant if bquant else cquant
        self.partOfKit = partOfKit
        self.sumb = sumb if sumb else sumn
        self.sumn = sumn
        self.active = True
        self.kitgoods = kitgoods


class CheckTotals(object):
    _Zero = Decimal('0.000')

    def __init__(self, check):
        self.check = check
        self._nextPosNum = self.check.getNextPosNum() 
        self.kitEntries = []
        self._catalogs = None
        self.pricetypes = None
        self.modifiedItems = {}

    def __repr__(self):
        return u'CheckTotals'

    def goodscodes(self):
        return uniq([item.code for item in self.vacantItemsIter()])

    def goodsItemsShort(self):
        res = {}
        for item in self.vacantItemsIter():
            if item.code not in res:
                res[item.code] = 0
            res[item.code] += item.cquant
        return res

    def catalogsItemsShort(self):
        """
        Построение словаря "каталог: количество товаров в чеке входящих в данный каталог"
        для отсечения наборов не встречающихся в чеке   
        """
        res = {}
        for item in self.vacantItemsIter():
            for catalog in item.catalogs:
                if catalog.catalogid not in res:
                    res[catalog.catalogid] = 0
                res[catalog.catalogid] += item.cquant
        return res
    
    def vacantItemsIter(self):
        for item in self._goodsItemsIter(self.check.getGoodsList()):
            if not item.partOfKit and not item.packingMode:
                yield item

    def kitItemsIter(self):
        for item in self._goodsItemsIter(self.check.getGoodsList()):
            if item.partOfKit:
                yield item
    
    def goodscq(self):
        resmap = {}
        for c, q in [ (item.code, item.bquant) for item in self.vacantItemsIter() ]:
            resmap[c] = resmap.get(c, self._Zero) + q
        return resmap

    def times(self, kit):
        min_t = None
        scq = self.goodscq()
        for c, q in kit.goodscq().items():
            t = int(scq.get(c, self._Zero) / q)
            min_t = min(t, coalesce(min_t, t))
        limit = kit.maxtimes
        result = coalesce(min_t, self._Zero)
        return min(limit, result) if limit else result

    def goodsItemsIter(self):  
        for item in self._goodsItemsIter(self.check.getGoodsList()):
            yield item
        
    def _goodsItemsIter(self, items):
        for item in items:
            if str(item.posnum) not in self.modifiedItems:
                yield DiscountObjectForPosition(item)
                continue
            for modifier in self.modifiedItems[str(item.posnum)]:
                if modifier.active:
                    yield DiscountObjectForPosition(item, modifier)

    def _subtractKitGoods(self, kg, times):
        entries = []
        code, cquant = kg.goodscode, kg.quantity * times
        for item in self.check.getGoodsList():
            if cquant <= self._Zero:
                break
            item = DiscountObjectForPosition(item)
            if item.code != code or item.packingMode:
                continue
            if str(item.posnum) not in self.modifiedItems:
                self.modifiedItems[str(item.posnum)] = [ DiscountKitModifiers(posnum=item.posnum, partOfKit=False, cquant=item.bquant, sumn=item.sumn) ]
            modifiers = [] 
            for modifier in self.modifiedItems[str(item.posnum)]:
                if (not modifier.active) or modifier.partOfKit:
                    continue
                
                # если количество товара в позиции  меньше или равно количеству в комплекте,
                # то позиция помечается как часть комплекта
                if modifier.cquant <= cquant:
                    modifier.partOfKit = True
                    modifier.kitgoods = kg
                    cquant -= modifier.cquant
                    entries.append(DiscountObjectForPosition(item, modifier))
                    break

                itemCquant = min(cquant, modifier.cquant)
                partOfKit = DiscountKitModifiers(posnum=self._nextPosNum, partOfKit=True, cquant=itemCquant, sumn=itemCquant * item.price, kitgoods=kg)
                modifiers.append(partOfKit)
                entries.append(DiscountObjectForPosition(item, partOfKit))
                self._nextPosNum += 1
                rquant = modifier.cquant - itemCquant
                if rquant > self._Zero:
                    modifiers.append(DiscountKitModifiers(posnum=self._nextPosNum, partOfKit=False, cquant=rquant, sumn=rquant * item.price))
                    self._nextPosNum += 1 
                modifier.active = False
                cquant -= itemCquant
                break
            self.modifiedItems[str(item.posnum)].extend(modifiers)
        return entries

    def subtract(self, kit, times):
        kitentries = []
        for kg in kit.kitgoods:
            goodsentries = self._subtractKitGoods(kg, times)
            ke = KitEntry(kit, goodsentries, kg, times)
            self.kitEntries.append(ke)
            kitentries.append(ke)
        return kitentries
    
    def kitSorted(self, kit):
        """ Сортировка записей набора """
        convolution = self.catalogsItemsShort()
        kit.kitgoods.sort(key=lambda x: (convolution[x.goodscode] if x.goodscode in convolution else 0) - x.quantity)
        return kit
    
    def distributionKit(self, kit):
        """ Поиск и выделение набора kit в чеке """
        numberKitApply = 0
        kitEntriesCopy = []
        while kit.maxtimes is None or numberKitApply < kit.maxtimes:
            kit_set = set(map(lambda x: x.goodscode, kit.kitgoods))
            kit = self.kitSorted(kit)
            modifiers = self.modifiedItems
            for kitentry in kit.kitgoods:
                catalog, cquant = kitentry.goodscode, kitentry.quantity
                for item in self.check.getGoodsList():
                    catalogList = map(lambda x: x.catalogid, item.getTmcObject().getCatalogsList())
                    # проверяем: может ли один товар быть 2-мя разными частями одного комплекта
                    if len(set(catalogList) & kit_set) > 1:
                        logger.warn("Goods '%s' is found in several kit" % item.code)
                    if cquant == 0:
                        break                   
                    if catalog not in catalogList:
                        continue                
                    item = DiscountObjectForPosition(item)
                    if str(item.posnum) not in self.modifiedItems:
                        modifiers[str(item.posnum)] = [DiscountKitModifiers(posnum=item.posnum, partOfKit=False, cquant=item.bquant, sumn=item.sumn)]
                    for modifier in modifiers[str(item.posnum)]:
                        if (not modifier.active) or modifier.partOfKit:
                            continue
                        # если количество товара в позиции  меньше или равно количеству в комплекте,
                        # то позиция помечается как часть комплекта
                        if modifier.cquant <= cquant:
                            modifier.partOfKit = True
                            modifier.kitgoods = kitentry
                            cquant -= modifier.cquant
                            kitEntriesCopy.append(KitEntry(kit, [DiscountObjectForPosition(item, modifier)], kitentry, 1))
                            break
                        # иначе создаем 2-а новых модификатора. Один - часть набора, второй - остаток не вошедший в набор
                        itemCquant = min(cquant, modifier.cquant)                
                        partOfKit = DiscountKitModifiers(posnum=self._nextPosNum, partOfKit=True, cquant=itemCquant, sumn=itemCquant * item.price, kitgoods=kitentry)
                        modifiers[str(item.posnum)].append(partOfKit)
                        kitEntriesCopy.append(KitEntry(kit, [DiscountObjectForPosition(item, partOfKit)], kitentry, 1))
                        self._nextPosNum += 1
                        rquant = modifier.cquant - itemCquant
                        if rquant > self._Zero:
                            modifiers[str(item.posnum)].append(DiscountKitModifiers(posnum=self._nextPosNum, partOfKit=False, cquant=rquant, sumn=rquant * item.price))
                            self._nextPosNum += 1 
                        # текущий модификатор блокируется
                        modifier.active = False       
                        cquant -= itemCquant
                        break
                # если хотя бы одна запись набора не может быть выделена,
                # то заканчиваем поиск записей текущего набора
                if cquant != 0:
                    return 
            # если набор был выделен полностью,
            # то сохраняем скидочные и структурные воздействия
            self.modifiedItems = modifiers
            self.kitEntries.extend(kitEntriesCopy)
            kitEntriesCopy = []
            numberKitApply += 1
         
    def branchImpacts(self):
        map = {}
        for key, modifiers in self.modifiedItems.items():
            for item in modifiers:
                if item.active and key != str(item.posnum):
                    map.setdefault(key, []).append({'qty': float(item.cquant), 'posnum': item.posnum, 'sum': float(item.sumb)})
        return map
        
    def getCatalogs(self):
        if self._catalogs is None:
            self._recalculateCatalogs()
        return self._catalogs
    
    def getCatalogsForPosition(self, item):
        catalogs = {}

        for catalog in item.getTmcObject().getCatalogsList():
            if catalog.catalogid not in catalogs:
                catalogs[catalog.catalogid] = {
                    'quant': Decimal('0.000'),
                    'sum': Decimal('0.00')
                }
            catalogs[catalog.catalogid]['quant'] += Decimal("%.03f" % item.cquant)
            catalogs[catalog.catalogid]['sum'] += Decimal("%.02f" % item.sumn)
            
        return catalogs
            
    def _recalculateCatalogs(self):
        self._catalogs = {}
        for item in self.check.getGoodsList():
            for catalog in item.getTmcObject().getCatalogsList():
                if catalog.catalogid not in self._catalogs:
                    self._catalogs[catalog.catalogid] = {
                        'quant': Decimal('0.000'),
                        'sum': Decimal('0.00')
                    }
                self._catalogs[catalog.catalogid]['quant'] += Decimal("%.03f" % item.cquant)
                self._catalogs[catalog.catalogid]['sum'] += Decimal("%.02f" % item.sumn)

    def getPricetypes(self):
        if self.pricetypes:
            return self.pricetypes
        self.pricetypes = uniq([item for item in self._pricetypesIter(self.check.getGoodsList())])
        return self.pricetypes

    def _pricetypesIter(self, items):    
        for item in items:
            yield item.pricetype if item.pricetype != 0 else None
