# -*- coding: utf-8 -*-

from simplediscount.checktotals import DiscountObjectForCheck, DiscountObjectForPosition, DiscountCardObject
from simplediscount.discauto import AutomaticDiscount
from simplediscount.disccard import DiscCard
from simplediscount.connection.connection import Session
from sqlalchemy import Table, MetaData, String, Numeric, Column
from sqlalchemy.sql import or_, func
from simplediscount.log_decorator import log_after
from enum import Enum
from simplediscount.sprices import Sprices
from simplediscount.discountconfig import DiscountConfig
from simplediscount.uniq import uniq
import datetime
import logging

session = Session()

logger = logging.getLogger('su.artix.dm')


class DiscountMode(Enum):
    AUTOMATIC = 'automatic'
    BYCARD = 'bycard'
    MANUAL = 'manual'
    FIXED = 'fixed'
    BYDISCKIT = 'bydisckit'
    BYPAYMENT = 'bypayment'


class AutomaticDiscountManager(object):
    discountSchemes = {}
    discountSchemesHeaders = {}
    catalogs = {}
    appliedSchemes = []
    appliedDisc = []

    def __init__(self, discConfig):
        pass

    def __repr__(self):
        return u'AutomaticDiscountManager'

    @log_after
    def loadSchemesFromDatabase(self, schemes, clean=True, onCheck=False, byCard=False):
        logger.debug('load a data for schemes: %s' % schemes)
        if clean == True:
            self.discountSchemesHeaders = {}
            self.discountSchemes = {}

        S = Sprices
        query = session.query(S)

        if onCheck:
            query = query.filter(S.oncheck == 1)
        else:
            query = query.filter(or_(S.oncheck == 0, S.oncheck == None))
        result = query.all()
        for rec in result:
            self.discountSchemesHeaders[str(rec.code)] = rec

        DA = AutomaticDiscount
        query = session.query(DA)
        schemes = self.discountSchemesHeaders.keys()
        query = query.filter(DA.shemecode.in_(schemes))
        query = query.order_by(DA.shemecode, DA.disccode)
        result = query.all()
        for rec in result:
            if not str(rec.shemecode) in self.discountSchemes:
                self.discountSchemes[str(rec.shemecode)] = []
            self.discountSchemes[str(rec.shemecode)].append(rec)

    @log_after
    def auto(self, object, check, schemeCode, byCatalog=False):
        logger.debug('trying automatic discounts')
        onCheck = isinstance(object, DiscountObjectForCheck)
        birthday = None
        discCardList = []
        for c in check.getCardsList():
            card = DiscountCardObject(c)
            card.disccard = DiscCard.select(card.card.number)
            if card.disccard:
                discCardList.append(card)
                logger.debug(u"card        : %s (%s)" % (card.disccard.cardnumber, card.disccard.name))
            else:
                logger.debug(u"no cards found with number '%s'" % card.card.number)

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug('scheme code     : %s' % (schemeCode))
            logger.debug('time            : %s' % (object.time))
            logger.debug('date            : %s' % (object.date))
            logger.debug('weekday         : %s' % (object.weekday))
            logger.debug('cquant          : %s' % (object.cquant))
            logger.debug('sumn            : %s' % (object.sumn))
        res = []
        if schemeCode and (
                (str(schemeCode) not in self.discountSchemesHeaders) or (str(schemeCode) not in self.discountSchemes)):
            return res
        catalogKeys = self.catalogs.keys()
        discList = []
        for disc in self.discountSchemes.values():
            discList.extend(disc)

        for disc in discList:
            disc.disccard = None
            if self.discountSchemesHeaders[str(disc.shemecode)].type == 1:
                if len(discCardList) == 0 or not disc.cardgroupcode:
                    continue
                findDisc = False
                for c in discCardList:
                    if c.disccard.groupcode == disc.cardgroupcode:
                        birthday = c.disccard.birthday if c.disccard.birthday else None
                        disc.disccard = c
                        findDisc = True
                        break
                if not findDisc:
                    continue
            if disc.catalogid:
                id = str(disc.catalogid)
                if id not in catalogKeys:
                    continue
                if disc.catalogquantbeg and (disc.catalogquantbeg > self.catalogs[id]['quant']):
                    continue
                if disc.catalogquantend and (disc.catalogquantend < self.catalogs[id]['quant']):
                    continue
                if disc.catalogsumbeg and (disc.catalogsumbeg > self.catalogs[id]['sum']):
                    continue
                if disc.catalogsumend and (disc.catalogsumend < self.catalogs[id]['sum']):
                    continue
            elif not schemeCode and not onCheck:
                continue
            if disc.birthdaymode and (disc.birthdaymode == 1) and (
                    birthday == None or ((object.date.day != birthday.day) or (object.date.month != birthday.month))):
                continue
            if (disc.date_beg and (object.date < disc.date_beg)) or (disc.date_end and (object.date > disc.date_end)):
                continue
            if (disc.time_beg and (object.time < disc.time_beg)) or (disc.time_end and (object.time > disc.time_end)):
                continue
            if (disc.days_beg and (disc.days_beg != 0) and (object.weekday < disc.days_beg)) or (
                    disc.days_end and (disc.days_end != 0) and (object.weekday > disc.days_end)):
                continue
            if (disc.quant_beg and (object.cquant < disc.quant_beg)) or (
                    disc.quant_end and (object.cquant > disc.quant_end)):
                continue
            if disc.sum_beg and (
                    disc.sum_beg > (object.sumn if (disc.rsummode == None) or (disc.rsummode == 0) else check.sumn)):
                continue
            if disc.sum_end and (
                    disc.sum_end < (object.sumn if (disc.rsummode == None) or (disc.rsummode == 0) else check.sumn)):
                continue
            res.append(disc)
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug('discount with code %s' % disc.disccode)
                logger.debug('    mode is %s' % disc.discmode)
                logger.debug('    type is %s' % disc.disctype)
                logger.debug('    val is %s' % disc.discval)
        return res

    @log_after
    def discountToItem(self, object, check):
        logger.debug('trying automatic discount for goods item : %s' % (object))
        schemeCode = object.pricetype if object.pricetype else None
        rates = self.auto(object, check, schemeCode)

        return [Discount(check, object, object, rate, rate.disccard,
                         DiscountMode.AUTOMATIC.value if rate.disccard == None else DiscountMode.BYCARD.value) for rate
                in rates]

    @log_after
    def discountToCheck(self, check):
        logger.debug('trying automatic discount for check: %s' % (check))
        rates = self.auto(check, check, None)

        return [Discount(check, check, check, rate, rate.disccard,
                         DiscountMode.AUTOMATIC.value if rate.disccard == None else DiscountMode.BYCARD.value) for rate
                in rates]


class Discount(object):
    check = None
    object = None
    rate = None
    card = None
    innerObject = None

    def __init__(self, check, object, innerObject, rate, card, mode):
        self.check = check
        self.object = object
        self.rate = rate
        self.card = card
        self.discountMode = mode
        self.innerObject = innerObject
        self.absoluteValue = self.rate.appliedValue(self.innerObject)
        self.deniedPositions = []

    def toDict(self):
        return {
            'discountRate': {'type': self.rate.type, 'value': float(self.rate.value)},
            'discountMode': self.discountMode,
            'absoluteValue': float(self.absoluteValue),
            'positionNumber': self.innerObject.posnum,
            'campaign': {'id': self.rate.scheme.code, 'name': self.rate.scheme.name},
            'discount': {'id': self.rate.disccode, 'name': self.rate.name, 'value': float(self.rate.value)},
            'cards': [self.card.dump()] if self.card else [],
            'deniedPositions': self.deniedPositions
        }

    def __repr__(self):
        if not self.card:
            return u'Discount (check = %s, object = %s, innerObject = %s, rate = %s, mode = %s)' % (
            self.check, self.object, self.innerObject, self.rate, self.discountMode)
        else:
            return u'Discount (check = %s, object = %s, innerObject = %s, rate = %s, mode = %s, card = %s)' % (
            self.check, self.object, self.innerObject, self.rate, self.discountMode, self.deniedPositions, self.card)


class DiscountCollection(object):
    class Over:
        def __call__(self, all, next):
            return [next]

    class Deny:
        def __call__(self, all, next):
            return [(all + [next])[0]]

    class Min:
        def __call__(self, all, next):
            return [min(all + [next], key=lambda x: x.absoluteValue)]

    class Max:
        def __call__(self, all, next):
            return [max(all + [next], key=lambda x: x.absoluteValue)]

    class First:
        def __call__(self, all, next):
            return [(all + [next])[0]]

    class Last:
        def __call__(self, all, next):
            return [next]

    class All:
        def __call__(self, all, next):
            return all + [next]

    class Add:
        def __call__(self, all, next):
            return all + [next]

    discounts = []

    def __init__(self, discConfig, discounts=None):

        self.discMode = discConfig[
            'SimpleDiscount.checkDiscRelation'].lower() if 'SimpleDiscount.checkDiscRelation' in discConfig else 'over'
        self.discPositionMode = discConfig[
            'SimpleDiscount.posDiscRelation'].lower() if 'SimpleDiscount.posDiscRelation' in discConfig else 'deny'
        self.discRelation = discConfig[
            'SimpleDiscount.posToCheckDiscRelation'].lower() if 'SimpleDiscount.posToCheckDiscRelation' in discConfig else 'all'

        logger.debug(
            u"discount collection with checkDiscRelation = '%s', posDiscRelation = '%s', posToCheckDiscRelation = %s created."
            % (self.discMode, self.discPositionMode, self.discRelation))
        if self.discMode == "over":
            self.checkMod = self.Over()
        elif self.discMode == "deny":
            self.checkMod = self.Deny()
        elif self.discMode == "all":
            self.checkMod = self.All()
        elif self.discMode == "max":
            self.checkMod = self.Max()
        elif self.discMode == "min":
            self.checkMod = self.Min()
        else:
            self.checkMod = self.Over()

        if self.discPositionMode == "over":
            self.posMod = self.Over()
        elif self.discPositionMode == "deny":
            self.posMod = self.Deny()
        elif self.discPositionMode == "all":
            self.posMod = self.All()
        elif self.discPositionMode == "min":
            self.posMod = self.Min()
        elif self.discPositionMode == "max":
            self.posMod = self.Max()
        else:
            self.posMod = self.Deny()

        self.discounts = []
        if discounts:
            self.extend(discounts)

    def __iter__(self):
        for discount in self.discounts:
            yield discount

    def __repr__(self):
        return 'DiscountCollection of %d items' % (len(self.discounts))

    def _appendDiscount(self, discount):
        if isinstance(discount.innerObject, DiscountObjectForCheck) and self.discRelation == u'deny':
            positions = uniq(
                [d.innerObject.posnum for d in self.discounts if isinstance(d.innerObject, DiscountObjectForPosition)])
            discount.deniedPositions = positions

        all = [d for d in self.discounts if d.innerObject == discount.innerObject]
        for d in all:
            self.discounts.remove(d)
        all = self.checkMod(all, discount) if isinstance(discount.innerObject, DiscountObjectForCheck) else self.posMod(
            all, discount)
        self.discounts.extend(all)

    def appendDiscount(self, check, object, innerObject, rate):
        discount = Discount(check, object, innerObject, rate)
        self._appendDiscount(discount)

    @log_after
    def extend(self, discounts):
        for disc in discounts:
            self._appendDiscount(disc)

    def __nonzero__(self):
        return bool(self.discounts)


class DiscountManager(object):

    def __init__(self, totals):
        self.totals = totals
        self.config = self.loadConfig()
        self.adm = AutomaticDiscountManager(self.config)

    def __repr__(self):
        return u'DiscountManager'

    @log_after
    def loadConfig(self):
        logger.debug('load config')
        discConfig = {}
        DC = DiscountConfig
        query = session.query(DC)
        result = query.all()
        for rec in result:
            logger.debug(u"load option %s: %s" % (rec.name, rec.value))
            discConfig[str(rec.name)] = rec.value

        return discConfig

    @log_after
    def discount(self, check):
        res = DiscountCollection(self.config)
        self.adm.loadSchemesFromDatabase(self.totals.getPricetypes(), True)
        self.adm.catalogs = self.totals.getCatalogs()
        checkObject = DiscountObjectForCheck(check)
        for object in self.totals.goodsItemsIter():
            self.adm.catalogs = self.totals.getCatalogsForPosition(object._position)
            res.extend(self.adm.discountToItem(object, checkObject))

        return res

    @log_after
    def discountOnCheck(self, check):
        res = DiscountCollection(self.config)
        self.adm.loadSchemesFromDatabase([], True, True)
        self.adm.catalogs = self.totals.getCatalogs()
        checkObject = DiscountObjectForCheck(check)
        res.extend(self.adm.discountToCheck(checkObject))

        return res

    @log_after
    def applyDiscRelation(self, discounts):
        if discounts.discRelation == u'max':
            sumCheck = sum(d.absoluteValue for d in discounts if isinstance(d.innerObject, DiscountObjectForCheck))
            sumPos = sum(d.absoluteValue for d in discounts if isinstance(d.innerObject, DiscountObjectForPosition))
            all = [d for d in discounts]
            for d in all:
                if (isinstance(d.innerObject, DiscountObjectForPosition) and sumPos < sumCheck) or (
                        isinstance(d.innerObject, DiscountObjectForCheck) and sumPos > sumCheck):
                    discounts.discounts.remove(d)
